
/*
 *  UnWindows 1.1.2
 *  change_sounds.c
 *
 *  Gary Ormsby
 *  Rensselaer Polytechnic Institute
 *  Fall 1991
 *
 *  This is the interface to change the sounds in the
 *  .un_sound file.  Sounds can be added, deleted, or
 *  modified.  Sounds are bound to windows or screen edges.
 *  A sound is bound to a window by entering the name of the
 *  window.  A sound is bound to a screen edge by entering
 *  Top_Border, Bottom_Border, Left_Border, Right_Border.
 *
 *  Minor revisions by Richard L. Kline
 *  December 1993
 *
 *  Further minor changes by Rich Burridge
 *  October 2001, January 2002, October 2002.
 *
 *  The program now uses the .un_sound file in the user's
 *  home directory rather than ~ormsbyg/proj/audio_specs.
 *  The "quit" button was moved from the far left to the far
 *  right of the row of buttons.
 *
 *  ---
 *  Modification history:
 *  1.1.2 - October 2001 - see ChangeLog file.
 *  1.1.1 - October 2001 - see ChangeLog file.
 *  1.1   - 16 May 94    - no changes.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Cardinals.h>
#include <math.h>
#include <X11/Shell.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/cursorfont.h>
#include <Xm/Protocols.h>

#include <sys/types.h>
#include <dirent.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "config.h"

#define CLOSEDIR (void) closedir
#define FCLOSE   (void) fclose
#define FPRINTF  (void) fprintf
#define PRINTF   (void) printf
#define SPRINTF  (void) sprintf
#define SSCANF   (void) sscanf
#define STRCAT   (void) strcat
#define STRCPY   (void) strcpy

#ifndef  AUDIO_PROG
#define  AUDIO_PROG      "/bin/audioplay"
#endif /*!AUDIO_PROG*/

#ifndef  SOUND_DIR
#define  SOUND_DIR      "sounds"
#endif /*!SOUND_DIR*/

static char *audio_prog;
static char audio_specs_file[128];
static char *ProgramName;

static int read_dir(char **, char *);
static void sort(char **, int);
static void make_text_sound(Boolean);
static void make_text_list(Widget, Boolean);
static void makepopup(Widget);
static void display_set_members(Widget, XEvent *, String *, Cardinal *);
static void display_sound(Widget, XEvent *, String *, Cardinal *);
static void play_sound(char *, char *);
static void play_sound_proc(Widget, XtPointer, XtPointer);
static void reset_dir_proc(Widget, XtPointer, XtPointer);
static void save_sound_proc(Widget, XtPointer, XtPointer);
static void quit_proc(Widget, XtPointer, XtPointer);
static void play_selection(Widget, XEvent *, String *, Cardinal *);
static void remove_sound(int);
static void read_data();
static void move_pt(Widget, int);
static void change_width(Widget, char *);
static int read_sound_file();
static char *find_sound(char *);
static char *num_to_str(int);
static int str_to_num(char *);

extern void playit(char *, char **);

static void delete_sound_proc(Widget, XtPointer, XtPointer);
static void add_sound_proc(Widget, XtPointer, XtPointer);
static void Usage();

static XtAppContext appCtx;

static Atom XA_WM_DELETE_WINDOW;

static String setnametrans =
    "<Motion>:  select-start()\n\
    <Btn1Down>: select-end(PRIMARY, CUT_BUFFER0) display_set_members()\n\
    <Btn2Down>: select-end(PRIMARY, CUT_BUFFER0) play_selection()";

static String setsoundtrans =
    "<Motion>:  select-start()\n\
    <BtnDown>:  select-end(PRIMARY, CUT_BUFFER0) display_sound()";

static XtActionsRec sound_actionTable [] = {
    { "display_sound", display_sound },
};

static XtActionsRec actionTable [] = {
    { "display_set_members", display_set_members },
    { "play_selection", play_selection }
};

static XtTranslations trans;

static int memselect[] = {
    XawselectLine, 
    XawselectNull, 
    XawselectNull, 
    XawselectNull, 
    XawselectNull, 
    XawselectNull
};

static int dialogselect[] = {
    XawselectNull
};

typedef struct Sound_struct Sound_struct;

struct Sound_struct {
    String window;
    String directory;
    String sound;
    String volume;
    String index;

    String old_window;
    String old_directory;
    String old_sound;
    String old_volume;
    String old_index;
};

#define INITIAL_LEVEL   0.5
#define MAXFILES        300
#define MAXSOUNDS       75
#define MAXSOUNDSLENGTH 2
#define WINDLENGTH      45
#define DIRLENGTH       255
#define VOLLENGTH       3
#define LINELENGTH      500
#define INDEXLENGTH     3

#define MAXVOLUME       100
#define NZ              NULL,ZERO

static Widget quit;
static Widget add_sound;
static Widget delete_sound;
static Widget save_sounds;

static Widget text_list;
static Widget text_sound;
static Widget sound_dir;
static Widget main_box;

static Widget popup;                 /* the popup parameter window */
static Widget index_dialog;
static Widget window_dialog;
static Widget sound_dialog;
static Widget volume_dialog;
static Widget play;

static String label_font_string = "-b&h-lucidatypewriter-bold-r-normal-sans-*-350-*-*-*-*-iso8859-*";

static Boolean delete_snd = False;

static Sound_struct *sounds[MAXSOUNDS];
static int num_sounds;
static String sound_text_string;

static Widget topLevel = NULL;

static String fallback_resources[] = {
    "*font: -b&h-lucidatypewriter-bold-r-normal-sans-*-350-*-*-*-*-iso8859-*",
    "*resizable:                        True",
    "*main_box.height:                  550",
    "*main_box.width:                   1000",
    "*main_box*label_example*fromVert:  scroll_bar",
    "*main_box*label_example*resizable: True",
    "*label_example*x:                  530",
    "*label_example*y:                  20",

    "*input:                            True",
    "*showGrip:                         off",
    "*text_sound*preferredPaneSize:     200",
    "*text_sound*editType:              read",
    "*text_sound*scrollVertical:        always",
    "*text_sound*scrollHorizontal:      never",
    "*text_sound*autoFill:              on",

    "*thickness:                        25",
    "*scroll_bar*x:                     380",
    "*scroll_bar*y:                     75",
    "*scroll_bar*height:                80",
    "*scroll_bar*width:                 500",

    "*bottomform.borderWidth:           0",

    "*popdown.fromHoriz:                ok",

    "*index_dialog.fromVert:            sound_dir",
    "*window_dialog.fromVert:           index_dialog",
    "*sound_dialog.fromVert:            window_dialog",
    "*volume_dialog.fromVert:           sound_dialog",

    "*text_list.fromHoriz:              dialogform",
    "*text_list.horizDistance:          350",
    "*text_list*preferredPaneSize:      200",
    "*text_list*editType:               read",
    "*text_list*scrollVertical:         always",
    "*text_list*scrollHorizontal:       never",
    "*text_list*autoFill:               on",

    "*appDefaultsVersion:               1",

    "XFontSel.cursor:                   left_ptr",
    "*allowShellResize:                 true",

    "*quitButton.Label:                 quit",
    "*quitButton.left:                  chainLeft",
    "*quitButton.right:                 chainLeft",

    "*selectButton.Label:               select",
    "*selectButton.fromHoriz:           quitButton",
    "*selectButton.left:                chainLeft",
    "*selectButton.right:               chainLeft",

    NULL,
};

static Display *dpy;           /* The X11 display */

static char *main_dir;
static char *Empty_String      = "Sound File is Empty";

static char *text_str          = NULL;
static int  max_text_pixel_len = 0;
static int  curindex           = 0;

static char *main_items[MAXFILES];
static int  num_main;


/* Used to respond to DELETE_WINDOW protocol request. */

static void
WindowMessageCB(Widget w, XtPointer client_data, XtPointer call_data)
{
    exit(0);
}


int
main(int argc, char **argv)
{
    char *home;
    int i;
    Arg arg[10];

    audio_prog = strdup(AUDIO_PROG);
    main_dir   = strdup(SOUND_DIR);

    ProgramName = argv[0];

    for (i = 1; i < argc; i++) {
        char *arg = argv[i];

        if (arg[0] == '-') {         /* process command line switches */
            switch (arg[1]) {
                case 'a':            /* -a <audio prog> */
                    if (++i >= argc) {
                        Usage();
                    }   
                    audio_prog = strdup(argv[i]);
                    continue;

                case 's':            /* -s <sound directory> */
                    if (++i >= argc) {
                        Usage();
                    }   
                    main_dir = strdup(argv[i]);
                    continue;

                default:
                    Usage();
            }
        } else {                     /* any command line parameters */
            Usage();
        }
    }

    if ((home = getenv("HOME")) != NULL) {
        SPRINTF(audio_specs_file, "%s/.un_sound", home);
    } else {
        STRCPY(audio_specs_file, "/.un_sound");
    }

    topLevel = XtAppInitialize(&appCtx, "XEyes2", NULL, 0,
                               &argc, argv, fallback_resources, NULL, 0);
    dpy = XtDisplay(topLevel);
    main_box = XtCreateManagedWidget("main_box",
                                     formWidgetClass, topLevel,
                                     NULL, 0);

    /* The ADD SOUND button */

    XtSetArg(arg[0], XtNx, 15);
    XtSetArg(arg[1], XtNy, 37);
    add_sound = XtCreateManagedWidget("Add Sound", 
                                      commandWidgetClass, main_box, 
                                      arg, 2);

    /* The DELETE SOUND button */

    XtSetArg(arg[0], XtNx, 15);
    XtSetArg(arg[1], XtNy, 147);
    XtSetArg(arg[2], XtNfromHoriz, add_sound);
    delete_sound = XtCreateManagedWidget("Delete Sound", 
                                         commandWidgetClass, main_box, 
                                         arg, 3);

    /* This is a window that allows the user to change the sound (and
     * its volume) bound an to an individual window (or screen edge)
     */

    makepopup(add_sound);

    /* The SAVE button */

    XtSetArg(arg[0], XtNx, 15);
    XtSetArg(arg[1], XtNy, 257);
    XtSetArg(arg[2], XtNfromHoriz, delete_sound);
    save_sounds = XtCreateManagedWidget("Save Sounds", 
                                        commandWidgetClass, main_box, 
                                        arg, 3);

    /* The QUIT button */

    XtSetArg(arg[0], XtNfromHoriz, save_sounds);
    quit = XtCreateManagedWidget("Quit", 
                                 commandWidgetClass, main_box, 
                                 arg, 1);

    XtAddCallback(quit,         XtNcallback, quit_proc,         NULL);
    XtAddCallback(save_sounds,  XtNcallback, save_sound_proc,   NULL);
    XtAddCallback(delete_sound, XtNcallback, delete_sound_proc, NULL);
    XtAddCallback(add_sound,    XtNcallback, add_sound_proc,    NULL);

    /* Read the audio_specs file */
    num_sounds = read_sound_file();

    /* Create the list of windows and their sounds */
    make_text_sound(True);                  /* Create text widget. */

    XtRealizeWidget(topLevel);

    XA_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(topLevel, XA_WM_DELETE_WINDOW,
                            WindowMessageCB, (caddr_t) 0);

    XtAppMainLoop(appCtx);

    return(0);
}


/* num_to_str - converts an int to a string.
 *
 * global variables - none
 *
 * inputs  - num - the int to convert
 * outputs - str - destination of the converted integer
 */

static char *
num_to_str(int num)
{
    char str[9];
    int i = 0, index = 0;
    int power = 10000;

    if (num == 0) {
        return("0");
    }

    while (power != 0) {
        i = num / power;
        num = num - i * power;
        if (index > 0 || i != 0) {
            str[index] = toascii(i + '0');
            index++;
        }
        power = power / 10;
    }
    str[index] = '\0';

    return(strdup(str));
}


/* str_to_num - converts a string to a int
 *
 * global variables - none
 *
 * inputs  - str - the character string to convert
 * outputs - returns an integer
 */

static int
str_to_num(char *str)
{
    int i, n;

    n = 0;
    for (i = 0; str[i] >= '0' && str[i] <= '9'; ++i) {
        n = 10 * n + str[i] - '0';
    }

    return (n);
}


static int
ConvertColor(Widget w, char *color_name)
{
    XrmValue from, to;

    from.size = strlen(color_name) + 1;
    from.addr = color_name;

/* This conversion accepts a colorname from rgb.txt, or a #rrrgggbbb
 * rgb color definition, or the special toolkit strings "XtDefaultForeground"
 * and "XtDefaultBackground".
 */

    XtConvert(w, XtRString, (XrmValuePtr) &from, XtRPixel, (XrmValuePtr) &to);
    if (to.addr == NULL) {
        return(-1);
    }

    return((int) *((Pixel *) to.addr));
}


/* This procedure displays the popup window. The user can enter the name of 
 * a window (or screen edge), bind a sound to it, and choose the volume with 
 * which to play the sound at.
 */

/*ARGSUSED*/
static void
add_sound_proc(Widget widget, XtPointer closure, XtPointer call_data)
{
    Arg args[10];
    char *empty_default = "        ";
    char *volume_default = "75";
    Sound_struct *p;

    p = (Sound_struct *) malloc(sizeof(Sound_struct));

    p->window = (char *) malloc(strlen(empty_default));
    STRCPY(p->window, empty_default);

    p->directory = (char *) malloc(strlen(main_dir));
    STRCPY(p->directory, main_dir);

    p->volume = (char *) malloc(strlen(volume_default));
    STRCPY(p->volume, volume_default);

    p->sound = (char *) malloc(strlen(empty_default));
    STRCPY(p->sound, (char *) empty_default);

    p->index = (char *) malloc(strlen(num_to_str(num_sounds)));
    STRCPY(p->index, num_to_str(num_sounds));

    sounds[num_sounds] = p;

    curindex = num_sounds;

    XtSetArg(args[0], XtNvalue, num_to_str(num_sounds));
    XtSetValues(index_dialog, args, 1);
    num_sounds++;

    XtSetArg(args[0], XtNvalue, empty_default);
    XtSetValues(window_dialog, args, 1);
    change_width(window_dialog, empty_default);

    XtSetArg(args[0], XtNvalue, empty_default);
    XtSetValues(sound_dialog, args, 1);
    change_width(sound_dialog, empty_default);

    XtSetArg(args[0], XtNvalue, volume_default);
    XtSetValues(volume_dialog, args, 1);
    XtPopup(popup, XtGrabExclusive);
}


/*ARGSUSED*/
static void
delete_sound_proc(Widget widget, XtPointer closure, XtPointer call_data)
{
    Arg     args[10];
    String  selected_color = "gray71";
    int     pixel;
    int     i;

    if (num_sounds == 0) {
        return;
    }

    if ((pixel = ConvertColor(widget, selected_color)) == -1) {
        FPRINTF(stderr, "Cannot convert color: %s\n", selected_color);
    } else {
        i = 0;
        XtSetArg(args[i], XtNbackground, (Pixel) pixel); 
        i++;
        XtSetValues(widget, args, i);
    }

    delete_snd = True;
}


/* This is a generic procedure to create dialog widgets. */

static void
create_dialog(Widget *w, Widget *parent, Widget neighbor, 
              char obj_class[], char value[], char label[], Boolean oneline)
{
    int n;
    Arg args[10];

    Widget tw;                     /* the Dialog widget's value */
    Widget lw;                     /* the Dialog widget's label */

    /* The following variables are used to override the dialog widget's
     * return key.  Instead of the cursor going to the next line, the
     * return key signifies end of input and calls read_data().
     */

    XtTranslations trans;           /* translations */

    String newTranslations = "<Key>Return: read_data()";

    static XtActionsRec actionTable[1] = { 
        { "read_data", read_data }
    };

    XtAppAddActions(appCtx, actionTable, XtNumber(actionTable));

    trans = XtParseTranslationTable(newTranslations);

    n = 0;
    XtSetArg(args[n], XtNvalue, (XtArgVal) value); 
    n++;
    XtSetArg(args[n], XtNlabel, (XtArgVal) label); 
    n++;
    XtSetArg(args[n], XtNborderWidth, 0); 
    n++;
    XtSetArg(args[n], XtNfromVert, neighbor); 
    n++;
    XtSetArg(args[n], XtNselectTypes, (XtArgVal) dialogselect); 
    n++;
    *w = XtCreateManagedWidget(obj_class, dialogWidgetClass, *parent, args, n);

    /* Override the return key in the dialog box's AsciiText child. */

    lw = XtNameToWidget(*w, "*label");
    tw = XtNameToWidget(*w, "*value");
    if (tw !=  NULL) {
        n = 0;
        XtSetArg(args[n], XtNresize, XawtextResizeWidth); 
        n++;
        XtSetArg(args[n], XtNborderWidth, 0); 
        n++;
        if (oneline) {
            XtSetArg(args[n], XtNfromHoriz, lw); 
            n++;
            XtSetArg(args[n], XtNfromVert, None); 
            n++;
        }
        XtSetArg(args[n], XtNfromHoriz, lw); 
        n++;
        XtSetArg(args[n], XtNfromVert, None); 
        n++;
        XtSetValues(tw, args, n);
        XtOverrideTranslations(tw, trans);
    }
}


/* Verifies that the user entered a number for the index. */

static int
chk_index(Widget *widget, char *str)
{
    Arg args[3];
    Boolean error = FALSE;
    int i;

    if (str_to_num(str) > MAXSOUNDS) {
        error = TRUE;
        return(!error);
    }

    for (i = 0; i < strlen(str); i++) {
        if (!isdigit((int) str[i]) ) {
            XtSetArg(args[0], XtNvalue, (XtArgVal) sounds[curindex]->index);
            XtSetValues(*widget, args, 1);
            error = TRUE;
            break;
        }
    }
    if (!error && str_to_num(str) >= num_sounds) {
        STRCPY(str, num_to_str(num_sounds-1));
    }

    return(!error);
}


/* Verifies that the user entered a valid number for the volume. */

static int
chk_volume(Widget *widget, char *str[])
{
    Arg args[3];
    Boolean error = FALSE;
    int i;
    int index;

    if (strlen(*str) == 0) {
        error = TRUE;
        return(!error);
    }

    index = str_to_num(XawDialogGetValueString(index_dialog)); 
    for (i = 0; i < strlen(*str); i++) {
        if (!isdigit((int) (*str)[i])) {
            XtSetArg(args[0], XtNvalue, (XtArgVal) sounds[index]->volume);
            XtSetValues(*widget, args, 1);
            move_pt(*widget, strlen(sounds[index]->volume));
            error = TRUE;
            break;
        }
    }
    if (!error && (str_to_num(*str) < 0 || str_to_num(*str) > MAXVOLUME)) {
        XtSetArg(args[0], XtNvalue, (XtArgVal) sounds[index]->volume);
        XtSetValues(*widget, args, 1);
        move_pt(*widget, strlen(sounds[index]->volume));
        error = TRUE;
    }

    return(!error);
}


/* Verifies that a valid filename was entered for the sound. */

static int
chk_sound(char *str[])
{
    Boolean allspaces = TRUE;
    int i;
    char c;

    for (i = 0; i < strlen(*str); i++) {
        c = (char) ((*str)[i]);
        if (!isspace((int) c)) {
            allspaces = FALSE;
            break;
        }
    }

    return(!allspaces);
}


/* Read the data from the dialog widgets in the popup window. */

static void
read_data()
{
    String index_str  = XawDialogGetValueString(index_dialog);
    String window_str = XawDialogGetValueString(window_dialog);
    String sound_str  = XawDialogGetValueString(sound_dialog);
    String volume_str = XawDialogGetValueString(volume_dialog);

    if (chk_index(&index_dialog, index_str)) {
        if (str_to_num(sounds[curindex]->index) != str_to_num(index_str)) {
            if (str_to_num(sounds[curindex]->index) < num_sounds) {
                Sound_struct *temp = sounds[str_to_num(index_str)];

                sounds[str_to_num(index_str)] = sounds[curindex]; 
                sounds[curindex] = temp;
                STRCPY(sounds[curindex]->index, num_to_str(curindex));
                STRCPY(sounds[str_to_num(index_str)]->index, index_str);    
            }
        }
    }

    STRCPY(sounds[str_to_num(index_str)]->window, window_str);

    if (chk_volume(&volume_dialog, &volume_str)) {
        STRCPY(sounds[str_to_num(index_str)]->volume, volume_str);
    }

    if (chk_sound(&sound_str)) {
        int index = str_to_num(index_str);

        free(sounds[index]->sound);

        sounds[index]->sound = NULL;
        sounds[index]->sound = (char *) calloc(strlen(sound_str), 1);
        STRCPY(sounds[index]->sound, sound_str);
    }

    free(sounds[str_to_num(index_str)]->directory);
    sounds[str_to_num(index_str)]->directory = (char *) 
        calloc(strlen(main_dir) + 1 + 
               strlen(sounds[str_to_num(index_str)]->sound), 1);

    STRCPY(sounds[str_to_num(index_str)]->directory, main_dir);
    STRCAT(sounds[str_to_num(index_str)]->directory, "/");
    STRCAT(sounds[str_to_num(index_str)]->directory,
           sounds[str_to_num(index_str)]->sound);
}


/* Pop the popup widget down. */

/*ARGSUSED*/
static void
pdown(Widget widget, XtPointer closure, XtPointer call_data)
{
    read_data();
    make_text_sound(False);
    XtPopdown((Widget) closure);
}


/* Make the popup window that allows the user to manipulate a window's 
 * sound and the volume.
 */

static void
makepopup(Widget parent)
{
    Arg arg[10];
    Widget pbox, popdown;
    Widget reset_dir;
    Cardinal n;

    /* Set the size of the window */
    n = 0;
    XtSetArg(arg[n], XtNheight, 800);
    n++;
    XtSetArg(arg[n], XtNwidth, 800);
    n++;

    /* Create the popup widget */
    popup = XtCreatePopupShell("prompt", 
                               transientShellWidgetClass, parent,    
                               NULL, 0);

    /* Create a box to hold everything */
    n = 0;
    XtSetArg(arg[n], XtNorientation, "vertical");
    n++;
    pbox = XtCreateManagedWidget("pbox",
                                 formWidgetClass, popup, 
                                 arg, n);

    popdown = XtCreateManagedWidget("Done", 
                                    commandWidgetClass, pbox, 
                                    NULL, 0);
    XtAddCallback(popdown, XtNcallback, pdown, (caddr_t) popup);

    n = 0;
    XtSetArg(arg[n], XtNfromHoriz, popdown); 
    n++;
    play = XtCreateManagedWidget("Play Sound",
                                 commandWidgetClass, pbox,
                                 arg, n);

    XtAddCallback(play, XtNcallback, play_sound_proc, NULL);

    create_dialog(&sound_dir, &pbox, popdown, "sound_dir", 
                  main_dir, "Directory: ", True);

    n = 0;
    XtSetArg(arg[n], XtNfromVert, sound_dir); 
    n++;
    reset_dir = XtCreateManagedWidget("Read Directory",
                                      commandWidgetClass, pbox,
                                      arg, n);

    XtAddCallback(reset_dir, XtNcallback, reset_dir_proc, NULL);

    create_dialog(&index_dialog, &pbox, reset_dir, "index_dialog", 
                  "", "Index:  ", False);

    create_dialog(&window_dialog, &pbox, index_dialog, "window_dialog",
                  "", "Window: ", False);

    create_dialog(&sound_dialog, &pbox, window_dialog, "sound_dialog",
                  "", "Sound:  ", False);

    create_dialog(&volume_dialog, &pbox, sound_dialog, "volume_dialog", 
                  "", "Volume: ", False);

    num_main = read_dir(main_items, main_dir);

    /* Sort the sounds. */
    sort(main_items, num_main);

    /* Put the sounds in a text widget so they can be viewed and selected. */
    make_text_list(pbox, True);
}


/* Resets the directory where the sound files are. */

/*ARGSUSED*/
static void
reset_dir_proc(Widget w, XtPointer closure, XtPointer call_data)
{
    int    i;

    for (i = 0; i< num_main; i++) {
        free(main_items[i]); 
    }
    free(text_str);

    STRCPY(main_dir, XawDialogGetValueString(sound_dir));

    num_main = read_dir(main_items, main_dir);
    sort(main_items, num_main);

    make_text_list(XtParent(sound_dir), False);
}


/* Places windows, their sounds, and their volumes in a text widget for 
 * selection.
 */

static void
make_text_list(Widget parent, Boolean construct_text_widget)
{
    int i;
    Arg arg[10];
    XFontStruct *font;

    font = (XFontStruct *) XLoadQueryFont(dpy, label_font_string);

    if (font == NULL) {
        PRINTF("font is null\n");
    }

    text_str = NULL;
    for (i = 0; i < num_main; i++) {
        char *skip1 = ".";
        char *skip2 = "..";
        int  width;
     
        if (strcmp(main_items[i], skip1) == 0) {
            continue;
        }
        if (strcmp(main_items[i], skip2) == 0) {
            continue;
        }

        if (text_str == NULL) {
            text_str = (char *) malloc(strlen(main_items[i]) + 1);
            STRCPY(text_str, main_items[i]);
        } else {
            text_str = (char *) realloc(text_str, strlen(text_str) +
                                                  strlen(main_items[i]) + 1);
            STRCAT(text_str, main_items[i]);
        }

        if (max_text_pixel_len < 
            (width = XTextWidth(font, main_items[i], strlen(main_items[i])))) {
            max_text_pixel_len = width;
        }

        STRCAT(text_str, "\n");
    }

    if (construct_text_widget) {
        i = 0;
        XtSetArg(arg[i], XtNx, 60);
        i++;
        XtSetArg(arg[i], XtNy, 15);
        i++;
        XtSetArg(arg[i], XtNheight, 500);
        i++;
        XtSetArg(arg[i], XtNwidth, max_text_pixel_len + 30);
        i++;
        XtSetArg(arg[i], XtNstring, text_str);
        i++;
        XtSetArg(arg[i], XtNselectTypes, (XtArgVal) memselect);
        i++;
        XtSetArg(arg[i], XtNfromVert, sound_dir);
        i++;
        XtSetArg(arg[i], XtNfromHoriz, index_dialog);
        i++;
        text_list = XtCreateManagedWidget("text_list", 
                                          asciiTextWidgetClass, parent,
                                          arg, i);

        XtAppAddActions(appCtx, sound_actionTable, 1);
        trans = XtParseTranslationTable(setsoundtrans);
        XtOverrideTranslations(text_list, trans);
    } else {
        XtUnmapWidget(text_list);
        i = 0;
        XtSetArg(arg[i],XtNstring, text_str);
        i++;
        XtSetValues(text_list, arg, i);
        XtMapWidget(text_list);
    }
}


/* Build the entry strings for the text widget. */

static void
make_text_sound(Boolean construct_text_widget)
{
    Widget text_sink;
    Arg arg[10];
    int i, cnt, *tabs;
    XFontStruct *font;

    int max_window = 0;
    int max_sound = 0;
    int max_index = 0;
    int max_volume = 0;

    font = (XFontStruct *) XLoadQueryFont(dpy, label_font_string);
    if (font == NULL) {
        PRINTF("font is null\n");
    }

    free(sound_text_string); 

    sound_text_string = NULL;

    if (num_sounds == 0) {
        sound_text_string = (char *) calloc(strlen(Empty_String), 1);
        STRCPY(sound_text_string, Empty_String);
    } else {
        for (i = 0; i < num_sounds; i++) {
            if (sound_text_string == NULL) {
                sound_text_string = (char *) calloc(strlen(sounds[i]->window) +
                                                    INDEXLENGTH +
                                                    strlen(sounds[i]->sound) +
                                                    strlen(sounds[i]->volume) +
                                                    1 + 1 + 1 + 1, 1);
            } else {
                sound_text_string = (char *) realloc(sound_text_string,
                                                     strlen(sound_text_string) +
                                                     INDEXLENGTH +
                                                     strlen(sounds[i]->window) +
                                                     strlen(sounds[i]->sound) +
                                                     VOLLENGTH + 1 + 1 + 1 + 1);
            }
      
            STRCAT(sound_text_string, sounds[i]->index);
            STRCAT(sound_text_string, "\t");
            STRCAT(sound_text_string, sounds[i]->window);
            STRCAT(sound_text_string, "\t");
            STRCAT(sound_text_string, sounds[i]->sound);
            STRCAT(sound_text_string, "\t");
      
            for (cnt = VOLLENGTH-strlen(sounds[i]->volume); cnt; cnt--) {
                STRCAT(sound_text_string, " ");
            }

            STRCAT(sound_text_string, sounds[i]->volume);
            STRCAT(sound_text_string, "\n");
      
            if (strlen(sounds[i]->window) > max_window) {
                max_window = strlen(sounds[i]->window);
            }

            if (strlen(sounds[i]->sound) > max_sound) {
                max_sound = strlen(sounds[i]->sound);
            }

            if (strlen(sounds[i]->index) > max_index) {
                max_index = strlen(sounds[i]->index);
            }

            if (strlen(sounds[i]->volume) > max_volume) {
                max_volume = strlen(sounds[i]->volume);
            }
        }
    }
    if ( construct_text_widget ) {
        i = 0;
        XtSetArg(arg[i], XtNy, 15);
        i++;
        XtSetArg(arg[i], XtNheight, 500);
        i++;
        XtSetArg(arg[i], XtNwidth, 800);
        i++;
        XtSetArg(arg[i], XtNstring, sound_text_string);
        i++;
        XtSetArg(arg[i], XtNresize, XawtextResizeWidth);
        i++;
        XtSetArg(arg[i], XtNselectTypes, (XtArgVal)memselect);
        i++;
        XtSetArg(arg[i], XtNfromVert, quit);
        i++;
        text_sound = XtCreateManagedWidget("text_sound", 
                                           asciiTextWidgetClass, main_box,
                                           arg, i);

        XtAppAddActions(appCtx, actionTable, 2);
        trans = XtParseTranslationTable(setnametrans);
        XtOverrideTranslations(text_sound, trans);
    } else {
        XtUnmapWidget(text_sound);
        i = 0;
        XtSetArg(arg[i], XtNstring, sound_text_string);
        i++;
        XtSetValues( text_sound, arg, i);
        XawTextUnsetSelection(text_sound);
        XtMapWidget(text_sound);
    }

    tabs = (int *) malloc(3 * sizeof(int));
    *tabs = INDEXLENGTH + 2;
    *(tabs+1) = max_window + 2 + *tabs;
    *(tabs+2) = max_sound + 2 + *(tabs+1);

    XtSetArg(arg[0], XtNtextSink, &text_sink);
    XtGetValues(text_sound, arg, ONE);
    if (text_sink == NULL) {
        FPRINTF(stdout, "textSink widget not found");
    } else {
        XawTextSinkSetTabs(text_sink, 3, tabs);
    }
}


/*
 * This is called when the DELETE SOUND buttton is pressed.
 * Subsequently pressing a button over an entry will remove it.
 */

static void
remove_sound(int index)
{
    String color = "white";
    Pixel pixel;
    Arg arg[4];

    if ((pixel = ConvertColor(delete_sound, color)) == -1) {
        FPRINTF(stderr, "Cannot convert color: %s\n", color);
    } else {
        int i = 0;

        XtSetArg(arg[i],XtNbackground, (Pixel) pixel); i++;
        XtSetValues(delete_sound, arg, i);

        free(sounds[index]->window);
        sounds[index]->window = NULL;
        free(sounds[index]->directory);
        sounds[index]->directory = NULL;
        free(sounds[index]->sound);
        sounds[index]->sound = NULL;
        free(sounds[index]->volume);
        sounds[index]->volume = NULL;
        free(sounds[index]->index);
        sounds[index]->index = NULL;
        free(sounds[index]);
        sounds[index] = NULL;

        for (i = index; i < num_sounds-1; i++) {
            sounds[i] = sounds[i+1];
            free(sounds[i]->index);
            sounds[i]->index = NULL;
            sounds[i]->index = (char *) calloc(strlen(num_to_str(i)), 1);
            STRCPY(sounds[i]->index,num_to_str(i));
        }
        num_sounds--;
        make_text_sound(False);
        delete_snd = False;
    }
}


/* Copy the selected sound from the text widget to the sound dialog box. */

/*ARGSUSED*/
static void
display_sound(Widget widget, XEvent *event, 
              String *params, Cardinal *num_params)
{
    char line[LINELENGTH];
    int n = LINELENGTH;
    long begin, end;
    Arg args[5];

    XawTextGetSelectionPos(text_list, &begin, &end);
    STRCPY(line,XFetchBytes(XtDisplay(widget), &n));

    XtSetArg(args[0], XtNvalue, (XtArgVal) line);
    XtSetValues(sound_dialog, args, 1);
    move_pt(sound_dialog, strlen(line));
    change_width(sound_dialog, line);

    read_data();
}


/* Move the caret to the end of the word in the dialog box. */

static void
move_pt(Widget widget, int pos)
{
    Widget text = XtNameToWidget(widget, "*value");

    if (text !=  NULL) {
        XawTextSetInsertionPoint(text, pos);
    }
}


/* Make the dialog box as long the text in it. */

static void
change_width(Widget widget, char str[])
{
    Arg args[4];
    Widget text = XtNameToWidget(widget, "*value");
    XFontStruct *font;

    XtSetArg(args[0], XtNfont, &font);
    XtGetValues(text, args, ONE);
    if (font == NULL) {
        PRINTF("font is null\n");
    } else {
        int width;

        width = XTextWidth(font, str, strlen(str));
        XtSetArg(args[0], XtNwidth, width + 10);
        XtSetValues(text, args, 1);
    }
}


/* When the mouse is clicked on an entry. */

/*ARGSUSED*/
static void
display_set_members(Widget w, XEvent *event, 
                    String *params, Cardinal *num_params)
{
    char line[LINELENGTH], str[INDEXLENGTH];
    int n = LINELENGTH;
    long begin, end;
    int index;
    Arg args[3];
    XFontStruct *font;

    font = (XFontStruct *) XLoadQueryFont(dpy, label_font_string);
    if (font == NULL) {
        PRINTF("font is null\n");
    }

    XawTextGetSelectionPos(text_sound, &begin, &end);
    STRCPY(line, XFetchBytes(XtDisplay(w), &n));

    if (!strcmp(line, Empty_String)) {
        return;
    }

    STRCPY(str, (char *) strtok(line, " \t"));
    curindex = index = str_to_num(str);

    if (delete_snd && begin != end) {
        remove_sound(index);
        return;
    }

    change_width(sound_dir, main_dir);
    move_pt(sound_dir, strlen(main_dir));

    XtSetArg(args[0], XtNvalue, sounds[index]->index);
    XtSetValues(index_dialog, args, 1);
    move_pt(index_dialog, strlen(sounds[index]->index));

    XtSetArg(args[0], XtNvalue, sounds[index]->window);
    XtSetValues(window_dialog, args, 1);
    move_pt(window_dialog, strlen(sounds[index]->window));
    change_width(window_dialog, sounds[index]->window);

    XtSetArg(args[0], XtNvalue, sounds[index]->sound);
    XtSetValues(sound_dialog, args, 1);
    move_pt(sound_dialog, strlen(sounds[index]->sound));
    change_width(sound_dialog, sounds[index]->sound);

    XtSetArg(args[0], XtNvalue, sounds[index]->volume);
    XtSetValues(volume_dialog, args, 1);
    move_pt(volume_dialog, strlen(sounds[index]->volume));

    XtPopup(popup, XtGrabExclusive);
}


/* Free the storage and quit. */

/*ARGSUSED*/
static void
quit_proc(Widget w, XtPointer cl, XtPointer ca)
{
    int i;

    for (i = 0; i < num_sounds; i++) {
        free(sounds[i]->window);
        free(sounds[i]->directory);
        free(sounds[i]->sound);
        free(sounds[i]->volume);
        free(sounds[i]->index);
        free(sounds[i]);
    }
    XtDestroyApplicationContext(XtWidgetToApplicationContext(w));
    exit(0);
}


/* Write the data to the audio_specs file. */

/*ARGSUSED*/
static void
save_sound_proc(Widget w, XtPointer closure, XtPointer call_data)
{
    int i;
    FILE *fp;

    if ((fp = fopen(audio_specs_file, "w")) == NULL) {
        FPRINTF(stderr, "unable to open %s\n", audio_specs_file);
    } else {
        for (i = 0; i < num_sounds; i++) {
            FPRINTF(fp, "%s\t%s\t%s\n", 
                    sounds[i]->window, sounds[i]->directory, sounds[i]->volume);
        }
        FCLOSE(fp);
    }
}


/* Call the play_it() routine to play the sound passed. */

static void
play_sound(char *volume, char *audio_file)
{
    char *argv[10];           /* Arguments */
    char *vol_str = "-v";     /* Used in file read */

    argv[0] = audio_prog;
    argv[1] = vol_str;
    argv[2] = volume;
    argv[3] = audio_file;     /* The sound to play. */
    argv[4] = NULL;
    playit(ProgramName, argv);
}


/* Get the parameters to play the chosen sound. */

/*ARGSUSED*/
static void
play_sound_proc(Widget w, XtPointer closure, XtPointer call_data)
{
    String sound_str = XawDialogGetValueString(sound_dialog);
    String volume_str = XawDialogGetValueString(volume_dialog);
    char *dir;

    if (chk_volume(&volume_dialog, &volume_str) &&
        chk_sound(&sound_str)) {

        dir = (char *) calloc(strlen(main_dir) + 1 + strlen(sound_str), 1);
        STRCPY(dir, main_dir);
        STRCAT(dir, "/");
        STRCAT(dir, sound_str);
        play_sound(volume_str, dir);
        free(dir);
    }
}


/* Play the currently selected sound. */

/*ARGSUSED*/
static void
play_selection(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    char line[LINELENGTH];
    int n = LINELENGTH;
    int index;
    char str[INDEXLENGTH];
    long begin, end;

    XawTextGetSelectionPos(text_sound, &begin, &end);
    STRCPY(line, XFetchBytes(XtDisplay(w), &n));

    if (!strcmp(line, Empty_String)) {
        return;
    }

    STRCPY(str, (char *) strtok(line, " \t"));
    STRCPY(str, (char *) strtok(line, " \t"));
    index = str_to_num(str);
    if (delete_snd  && begin != end) {
        remove_sound(index);
        return;
    }
    play_sound(sounds[index]->volume, sounds[index]->directory);
}


/* Parse the window, sound, and volume from the entry chosen in the 
 * text widget.
 */

static int
read_sound_file()
{
    FILE *fp;
    char wind[WINDLENGTH], dir[DIRLENGTH], vol[VOLLENGTH+1], line[LINELENGTH];
    int cnt = 0;
    Sound_struct *p;

    if ((fp = fopen(audio_specs_file, "r")) == NULL) {
        FPRINTF(stderr,"unable to open %s\n", audio_specs_file);
    } else {
        FPRINTF(stdout, "opened %s\n", audio_specs_file);
        while (fgets(line, 500, fp)) {
            SSCANF(line, "%s %s %s", wind, dir, vol);

            p = (Sound_struct *) malloc(sizeof(Sound_struct));

            p->window = (char *) malloc(strlen(wind));
            STRCPY(p->window, wind);

            p->directory = (char *) malloc(strlen(dir));
            STRCPY(p->directory, dir);

            p->volume = (char *) malloc(strlen(vol));
            STRCPY(p->volume, vol);

            p->sound = (char *) malloc(strlen(find_sound(dir)));
            STRCPY(p->sound, (char *) find_sound(dir));

            p->index = (char *) malloc(strlen(num_to_str(cnt)));
            STRCPY(p->index,num_to_str(cnt));

            sounds[cnt] = p;

            cnt++;
        }
    }
    FCLOSE(fp);

    return(cnt);
}


/* Parse the filename out of the entire pathname. */

static char *
find_sound(char *str)
{
    char slash = '/';
    char *index = NULL;

    index = rindex(str, slash);
    if (index == NULL) {
        return(NULL);
    }
    index++;

    return(index);
}


/* Read the sound filenames out of the specified directory. */

static int
read_dir(char *entries[], char *directory)
{
    int i = 0;
    DIR *dirp;
    struct dirent *dp;
    char *p;

    if ((dirp = opendir(directory)) == NULL) {
        return(-1);
    }
    for (i = 0, dp = readdir(dirp); dp != NULL; ++i, dp = readdir(dirp)) {
        if (i >= MAXFILES) {
            CLOSEDIR(dirp);
            return(-1);
        } else if ((p = (char *) malloc(strlen(dp->d_name))) == NULL) {
            CLOSEDIR(dirp);
            return(-1);
        } else {
            STRCPY(p, dp->d_name);
            entries[i] = p;
        }
    }
    CLOSEDIR(dirp);

    return(i);
}


static void
sort(char *entries[], int number)
{
    int gap, i, j;
    char *temp;

    for (gap = number / 2; gap > 0; gap /= 2) {
        for (i = gap; i < number; i++) {
            for (j= i - gap; j >= 0; j -= gap) {
                if (strcmp(entries[j], entries[j + gap]) <= 0) {
                    break;
                }
                temp = entries[j];
                entries[j] = entries[j+gap];
                entries[j + gap] = temp;
            }
        }
    }
}


/* Usage - Print out usage message and quit. */

static void
Usage()
{
    static char *help_message[] = {
        "where options include:",
        "  -a <audio program>    Alternate audio program to play sounds",
        "  -s <sound directory>  Alternate directory containing sound files",
        "",
        NULL
    }, **cpp;

    FPRINTF(stderr, "Usage:\n        %s [-option ...]\n\n", ProgramName);
    for (cpp = help_message; *cpp; cpp++) {
        FPRINTF(stderr, "%s\n", *cpp);
    }

    exit(1);
}
