
/*
 *  UnWindows 1.1.2
 *  Eyes.c
 *
 *  Richard Kline
 *  Rensselaer Polytechnic Institute
 *  16 May 1994
 *
 *  Minor revisions by Rich Burridge
 *  Sun Microsystems Inc.
 *  October 2001. 
 *
 *  This program defines the Eyes widget with the capability
 *  of changing colors based on the pointer location.
 *
 *  ---
 *  Modification history:
 *  1.1.1 - October 2001 - See ChangeLog file.
 *  1.1   - 16 May 94    - eliminated private color allocation, replacing
 *                         it with read/write color cells, allowing eyes
 *                         to instantly incorporate changes in settings window
 */

/* NOTE: New code has a comment of "N". */

#include <X11/Xos.h>
#include <stdio.h>
#include <stdlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Converters.h>
#include "EyesP.h"
#include <math.h>
#ifdef SHAPE
#include <X11/extensions/shape.h>
#endif /*SHAPE*/

#define offset(field)  XtOffset(EyesWidget, eyes.field)
#define goffset(field) XtOffset(Widget,     core.field)

static XtResource resources[] = {
    { XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
      goffset(width), XtRString, "150" },
    { XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
      goffset(height), XtRString, "100" },
    
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
      offset(puppixel0), XtRString, "Black" },
    { XtNoutline, XtCForeground, XtRPixel, sizeof(Pixel),
      offset(outline0), XtRString, "Black" },
    { XtNcenterColor, XtCBackground, XtRPixel, sizeof(Pixel),
      offset(center0), XtRString, "White" },

    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
      offset(puppixel1), XtRString, "Black" },
    { XtNoutline, XtCForeground, XtRPixel, sizeof(Pixel),
      offset(outline1), XtRString, "Black" },
    { XtNcenterColor, XtCBackground, XtRPixel, sizeof(Pixel),
      offset(center1), XtRString, "White" },

    { XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof(Boolean),
      offset(reverse_video), XtRString, "FALSE" },
    { XtNbackingStore, XtCBackingStore, XtRBackingStore, sizeof(int),
      offset(backing_store), XtRString, "default" },
#ifdef SHAPE
    { XtNshapeWindow, XtCShapeWindow, XtRBoolean, sizeof(Boolean),
      offset(shape_window), XtRString, "TRUE" },
#endif /*SHAPE*/
};

#undef offset
#undef goffset

#define NUM_EYES     2
#define EYE_X(n)     ((n) * 2.0)
#define EYE_Y(n)     (0.0)
#define EYE_OFFSET   (0.1)              /* padding between eyes */
#define EYE_THICK    (0.175)            /* thickness of eye rim */
#define BALL_WIDTH   (0.3)
#define BALL_PAD     (0.05)
#define EYE_WIDTH    (2.0 - (EYE_THICK + EYE_OFFSET) * 2)
#define EYE_HEIGHT   EYE_WIDTH
#define BALL_HEIGHT  BALL_WIDTH
#define BALL_DIST    ((EYE_WIDTH - BALL_WIDTH) / 2.0 - BALL_PAD)
#define W_MIN_X      (-1.0 + EYE_OFFSET)
#define W_MAX_X      (3.0 - EYE_OFFSET)
#define W_MIN_Y      (-1.0 + EYE_OFFSET)
#define W_MAX_Y      (1.0 - EYE_OFFSET)

extern EyesColorRec dis_colors[];   /* variable number of color-% pairs */
extern int num_colors;              /* number of color-% pairs */
extern int max_x;                   /* width of screen */
extern int max_y;                   /* height of screen */
extern Pixel boxpix[];              /* r/w color cell pixel values */

/* new defines */
#define WHITE 0
#define BLACK 1

#define TPointEqual(a, b)  ((a).x == (b).x && (a).y == (b).y)
#define XPointEqual(a, b)  ((a).x == (b).x && (a).y == (b).y)

static int delays[] = { 50, 100, 200, 400, 0 };

static void Initialize(), Realize(), Destroy();
static void Redisplay(), Resize ();
static Boolean SetValues();
static void repaint_window(EyesWidget);
static void draw_it(XtPointer, XtIntervalId *);
static void eyeBall(EyesWidget, GC, int);
static void eyeLiner(EyesWidget, Drawable, GC, GC, int);

static void ClassInitialize();

static EyesClassRec eyesClassRec = {
    { /* core fields */
      /* superclass */               &widgetClassRec,
      /* class_name */               "Eyes",
      /* size */                     sizeof(EyesRec),
      /* class_initialize */         ClassInitialize,
      /* class_part_initialize */    NULL,
      /* class_inited */             FALSE,
      /* initialize */               Initialize,
      /* initialize_hook */          NULL,
      /* realize */                  Realize,
      /* actions */                  NULL,
      /* num_actions */              0,
      /* resources */                resources,
      /* num_resources */            XtNumber(resources),
      /* xrm_class */                NULL,
      /* compress_motion */          TRUE,
      /* compress_exposure */        TRUE,
      /* compress_enterleave */      TRUE,
      /* visible_interest */         FALSE,
      /* destroy */                  Destroy,
      /* resize */                   Resize,
      /* expose */                   Redisplay,
      /* set_values */               SetValues,
      /* set_values_hook */          NULL,
      /* set_values_almost */        NULL,
      /* get_values_hook */          NULL,
      /* accept_focus */             NULL,
      /* version  */                 XtVersion,
      /* callback_private */         NULL,
      /* tm_table */                 NULL,
      /* query_geometry */           XtInheritQueryGeometry,
    }
};


static void
ClassInitialize()
{
    XtAddConverter(XtRString, XtRBackingStore, XmuCvtStringToBackingStore,
                   NULL, 0);
}


WidgetClass eyesWidgetClass = (WidgetClass) &eyesClassRec;

static XColor white;      /* color for the "whites of the eyes" */
static GC pupGC;          /* GC for the pupil */
static GC outGC;          /* GC for the rims of the eyes */
static GC centerGC;       /* GC for the center of the eyes */

static int color = BLACK; /* color of the eyes - this will change to the colors
                           * specified in the xeyes_specs file
                           */

/* Allocate colormap and colormap entries. */

static void 
do_colors()
{
    int i;

    /* parse and allocate the colors we want */
    for (i = 0; i < 10; i++) {
        dis_colors[i].xcolor.pixel = boxpix[i];
    }
}


/* Initialize the widget */

/* ARGSUSED */
static void 
Initialize(Widget greq, Widget gnew)
{
    EyesWidget w = (EyesWidget) gnew;
    XtGCMask valuemask;
    XGCValues myXGCV;
#ifdef SHAPE
    int shape_event_base, shape_error_base;
#endif

    do_colors();         /* get colors */

    /* set the colors if reverse video; these are the colors used:
     *
     *     background - paper          white
     *     foreground - text, ticks    black
     *     border - border             black (foreground)
     *
     * This doesn't completely work since the parent has already made up a 
     * border. Sigh.
     */

    if (w->eyes.reverse_video) {
        Pixel fg = w->eyes.puppixel0;
        Pixel bg = w->core.background_pixel;

        if (w->core.border_pixel == fg) {
            w->core.border_pixel = bg;
        }
        if (w->eyes.outline0 == fg) {
            w->eyes.outline0 = bg;
        }
        if (w->eyes.center0 == bg) {
            w->eyes.center0 = fg;
        }
        w->eyes.puppixel0 = bg;
        w->core.background_pixel = fg;

        if (w->core.border_pixel == fg) {
            w->core.border_pixel = bg;
        }
        if (w->eyes.outline1 == fg) {
            w->eyes.outline1 = bg;
        }
        if (w->eyes.center1 == bg) {
            w->eyes.center1 = fg;
        }
        w->eyes.puppixel1 = bg;
        w->core.background_pixel = fg;
    }

    myXGCV.foreground = w->eyes.puppixel0;
    myXGCV.background = w->core.background_pixel;
    valuemask         = GCForeground | GCBackground;
    w->eyes.pupGC0    = XtGetGC(gnew, valuemask, &myXGCV);
    myXGCV.foreground = w->eyes.puppixel1;
    myXGCV.background = w->core.background_pixel;
    valuemask         = GCForeground | GCBackground;
    w->eyes.pupGC1    = XtGetGC(gnew, valuemask, &myXGCV);

    myXGCV.foreground = w->eyes.outline0;
    valuemask         = GCForeground | GCBackground;
    w->eyes.outGC0    = XtGetGC(gnew, valuemask, &myXGCV);
    myXGCV.foreground = w->eyes.outline1;
    valuemask         = GCForeground | GCBackground;
    w->eyes.outGC1    = XtGetGC(gnew, valuemask, &myXGCV);

    myXGCV.foreground = w->eyes.center0;
    myXGCV.background = w->eyes.puppixel0;
    valuemask         = GCForeground | GCBackground;
    w->eyes.centerGC0 = XtGetGC(gnew, valuemask, &myXGCV);
    myXGCV.foreground = w->eyes.center1;
    myXGCV.background = w->eyes.puppixel1;
    valuemask         = GCForeground | GCBackground;
    w->eyes.centerGC1 = XtGetGC(gnew, valuemask, &myXGCV);

    w->eyes.update = 0;
    /* wait for Realize to add the timeout */
    w->eyes.interval_id = 0;

    w->eyes.pupil[0].x = w->eyes.pupil[1].x = -1000;
    w->eyes.pupil[0].y = w->eyes.pupil[1].y = -1000;

#ifdef SHAPE
    if (w->eyes.shape_window &&
        !XShapeQueryExtension(XtDisplay(w), &shape_event_base,
                                            &shape_error_base)) {
        w->eyes.shape_window = False;
    }
    w->eyes.shape_mask = 0;
    w->eyes.shapeGC = 0;
#endif /*SHAPE*/
}


/* The widget's Resize procedure */

static void 
Resize(Widget gw)
{
    EyesWidget w = (EyesWidget) gw;

#ifdef SHAPE
    XGCValues xgcv;
    Widget parent;
    int x, y;
#endif /*SHAPE*/

    if (XtIsRealized(gw)) {
        XClearWindow(XtDisplay(w), XtWindow(w));
        SetTransform(&w->eyes.t, 0, w->core.width, w->core.height, 0,
                     W_MIN_X, W_MAX_X, W_MIN_Y, W_MAX_Y);

#ifdef SHAPE
        if (w->eyes.shape_window) {
            w->eyes.shape_mask = XCreatePixmap(XtDisplay(w), XtWindow(w),
                                             w->core.width, w->core.height, 1);
            if (!w->eyes.shapeGC) {
                w->eyes.shapeGC = XCreateGC(XtDisplay(w), 
                                            w->eyes.shape_mask, 0, &xgcv);
            }
            XSetForeground(XtDisplay(w), w->eyes.shapeGC, 0);
            XFillRectangle(XtDisplay(w), w->eyes.shape_mask, w->eyes.shapeGC, 
                           0, 0, w->core.width, w->core.height);
            XSetForeground(XtDisplay(w), w->eyes.shapeGC, 1);
            eyeLiner(w, w->eyes.shape_mask, w->eyes.shapeGC, (GC) 0, 0);
            eyeLiner(w, w->eyes.shape_mask, w->eyes.shapeGC, (GC) 0, 1);
            x = y = 0;
            for (parent = (Widget) w; XtParent(parent); 
                                      parent = XtParent(parent)) {
                x += parent->core.x + parent->core.border_width;
                x += parent->core.y + parent->core.border_width;
            }
            XShapeCombineMask(XtDisplay(parent), XtWindow(parent),
                            ShapeBounding, x, y, w->eyes.shape_mask, ShapeSet);
            XFreePixmap(XtDisplay (w), w->eyes.shape_mask);
        }
#endif /*SHAPE*/
    }
}


/* The widget's Realize procedure. */

static void 
Realize(Widget gw, XtValueMask *valueMask, XSetWindowAttributes *attrs)
{
    EyesWidget w = (EyesWidget) gw;

    XtGCMask valuemask;                                                  /* N */
    XGCValues myXGCV;                                                    /* N */
    int i;                                                               /* N */

    if (w->eyes.backing_store != Always + WhenMapped + NotUseful) {
        attrs->backing_store = w->eyes.backing_store;
        *valueMask |= CWBackingStore;
    }
    XtCreateWindow(gw, (unsigned) InputOutput, (Visual *) CopyFromParent,
                   *valueMask, attrs);

    myXGCV.background = white.pixel;                                     /* N */
    valuemask = GCForeground | GCBackground;                             /* N */
    
    /* get GCs for all the colors */
    for (i = 0; i < 10; i++) {                                           /* N */
        myXGCV.foreground = dis_colors[i].xcolor.pixel;                  /* N */
        dis_colors[i].colorGC = XtGetGC((Widget) w, valuemask, &myXGCV); /* N */
    }                                                                    /* N */
    
    outGC = XCreateGC(XtDisplay(w), XtWindow(w), valuemask, &myXGCV);    /* N */
    centerGC = XCreateGC(XtDisplay(w), XtWindow(w), valuemask, &myXGCV); /* N */
    pupGC = XCreateGC(XtDisplay(w), XtWindow(w), valuemask, &myXGCV);    /* N */
    valuemask = GCForeground | GCBackground;                             /* N */
    
    XCopyGC(XtDisplay(w), w->eyes.outGC0,    valuemask, outGC);          /* N */
    XCopyGC(XtDisplay(w), w->eyes.centerGC0, valuemask, centerGC);       /* N */
    XCopyGC(XtDisplay(w), w->eyes.pupGC0,    valuemask, pupGC);          /* N */

    Resize (gw);

    /* Cast on draw_it() function should not be necessary */
    w->eyes.interval_id = XtAddTimeOut(delays[w->eyes.update], 
                                  (XtTimerCallbackProc) draw_it, (caddr_t) gw);
}


/* The widget's Destroy procedure. */

static void 
Destroy(Widget gw)
{
    EyesWidget w = (EyesWidget) gw;

    if (w->eyes.interval_id) {
        XtRemoveTimeOut(w->eyes.interval_id);
    }
    XtDestroyGC(w->eyes.pupGC0);
    XtDestroyGC(w->eyes.outGC0);
    XtDestroyGC(w->eyes.centerGC0);

    XtDestroyGC(w->eyes.pupGC1);
    XtDestroyGC(w->eyes.outGC1);
    XtDestroyGC(w->eyes.centerGC1);
}


/* The widget's Redisplay procedure. */

/* ARGSUSED */
static void 
Redisplay(Widget gw, XEvent *event, Region region)
{
    EyesWidget w;

    w = (EyesWidget) gw;
    w->eyes.pupil[0].x = -1000;
    w->eyes.pupil[0].y = -1000;
    w->eyes.pupil[1].x = -1000;
    w->eyes.pupil[1].y = -1000;
    repaint_window((EyesWidget) gw);
}


/* Compute the location of the pupil within the eye */

static TPoint 
computePupil(int num, TPoint mouse)
{
    double cx, cy, dist, angle, x, y, h, a, b, dx, dy;
    TPoint ret;

    dx = mouse.x - EYE_X(num);
    dy = mouse.y - EYE_Y(num);
    if (dx == 0 && dy == 0) {
        cx = EYE_X(num);
        cy = EYE_Y(num);
    } else {
        angle = atan2((double) dy, (double) dx);
        a = EYE_WIDTH / 2.0;
        b = EYE_HEIGHT / 2.0;
        h = hypot(b * cos (angle), a * sin(angle));
        x = a * b * cos(angle) / h;
        y = a * b * sin(angle) / h;
        dist = BALL_DIST * hypot(x, y);

        if (dist > hypot((double) dx, (double) dy)) {
            cx = dx + EYE_X(num);
            cy = dy + EYE_Y(num);
        } else {
            cx = dist * cos(angle) + EYE_X(num);
            cy = dist * sin(angle) + EYE_Y(num);
        }
    }
    ret.x = cx;
    ret.y = cy;
    return(ret);
}


/* Compute the location of the pupils within both eyes. */

static void 
computePupils(TPoint mouse, TPoint pupils[2])
{
    pupils[0] = computePupil(0, mouse);
    pupils[1] = computePupil(1, mouse);
}


/* Translate the pupils' coordinates in the eyes' widget to the RootWindow. */

static int 
find_coords(EyesWidget w,
            Position *root_x0 /* N */, Position *root_y0 /* N */,
            Position *root_x1 /* N */, Position *root_y1 /* N */)
{
    XtTranslateCoords((Widget) w,
                      Xx(w->eyes.pupil[0].x, w->eyes.pupil[0].y, &w->eyes.t),
                      Xy(w->eyes.pupil[0].x, w->eyes.pupil[0].y, &w->eyes.t),
                      root_x0, root_y0);

    XtTranslateCoords((Widget) w,
                      Xx(w->eyes.pupil[1].x, w->eyes.pupil[1].y, &w->eyes.t),
                      Xy(w->eyes.pupil[1].x, w->eyes.pupil[1].y, &w->eyes.t),
                      root_x1, root_y1);

    return(1);
}


/* Based on the cursor's distance from the eyes, determine which color the 
 * eyes should be.
 */

static void 
draw_dist(float percent, EyesWidget w)
{
    int i;

    for (i = 0; i < num_colors; i++) {
        if (percent >= dis_colors[i].eyes_loffset &&
            percent < dis_colors[i].eyes_uoffset)
        if (color != dis_colors[i].number) {
            color = dis_colors[i].number;
            eyeLiner(w, XtWindow(w), dis_colors[i].colorGC, 
                     w->eyes.centerGC0, 0);
            eyeLiner(w, XtWindow(w), dis_colors[i].colorGC, 
                     w->eyes.centerGC1, 1);
            eyeBall(w, dis_colors[i].colorGC, 0);
            eyeBall(w, dis_colors[i].colorGC, 1);
            break;
        }
    }
}


/* ARGSUSED */
static void 
draw_it(XtPointer client_data, XtIntervalId *id)
{
    EyesWidget   w = (EyesWidget) client_data;
    Window       rep_root, rep_child;
    int          rep_rootx, rep_rooty;
    unsigned int rep_mask;
    int          dx, dy;
    TPoint       mouse;
    Display      *dpy = XtDisplay(w);
    Window       win = XtWindow(w);
    TPoint       newpupil[2];
    XPoint       xnewpupil, xpupil;
    Position     root_x0, root_y0;     /* N */
    Position     root_x1, root_y1;     /* N */
    int          avg_x, avg_y;         /* N */
    float        dist,dist2;           /* N */
    float        percent = 0.0;        /* N */
    float        angle;                /* N */
    float        curs_angle;           /* N */
    float        corn_dist_low;        /* N */
    float        corn_dist_up;         /* N */
    float        min_cosine;           /* N */
    float        max_cosine;           /* N */
    float        cosine;               /* N */
    int          i;                    /* N */

    if (XtIsRealized((Widget) w)) {
        Bool same_screen = XQueryPointer(dpy, win, &rep_root, &rep_child,
                                         &rep_rootx, &rep_rooty, 
                                         &dx, &dy, &rep_mask);

        if (same_screen == True) {
            if (find_coords(w, &root_x0, &root_y0, &root_x1, &root_y1) != 0) {
                avg_x = ((int) root_x0 + (int) root_x1) / 2;
                avg_y = ((int) root_y0 + (int) root_y1) / 2;

                corn_dist_low = (float) hypot((double) abs(max_x - rep_rootx),
                                (double) abs(max_y - rep_rooty));

                corn_dist_up = (float) hypot((double) abs(rep_rootx),
                               (double) abs(rep_rooty));

                max_cosine = (max_y - avg_y) / 
                             ((float) hypot((double) abs(max_x - avg_x),
                                            (double) abs(max_y - avg_y)));

                min_cosine = (avg_y) / 
                             ((float) hypot((double) abs(avg_x),
                                            (double) abs(avg_y)));

                if (rep_rooty > avg_y) {
                    curs_angle = (float) acos((double) (rep_rootx / 
                                 (double) hypot((double) abs(rep_rootx), 
                                            (double) abs(max_y - rep_rooty))));

                    angle = (float) acos((double) (avg_x /
                            (double) hypot((double) abs(avg_x),
                                           (double) abs(max_y - avg_y))));
                } else {
                    curs_angle = (float) acos((double) ((double) (rep_rooty) /
                        (double) (hypot((double) abs((int)(max_x - rep_rootx)),
                                        (double) abs(rep_rooty)))));

                    angle = (float) acos((double) ((double) (avg_y) /
                        (double) (hypot((double) abs((int) (max_x - avg_x)),
                                        (double) abs(avg_y)))));
                }

                dist = (float) hypot((double) abs(rep_rootx - avg_x),
                                     (double) abs(rep_rooty - avg_y));

                if (curs_angle < angle) {
                    if ((max_y - rep_rooty) / corn_dist_low > max_cosine) {
                        cosine = (rep_rootx - avg_x) / dist;
                        dist2 = fabs(((max_x - avg_x) / cosine) - dist);
                        percent = dist / (dist + dist2);    
                    } else {
                        cosine = (rep_rooty - avg_y) / dist;
                        dist2 = fabs(((max_y - avg_y) / cosine) - dist);
                        percent = dist / (dist + dist2);
                    }
                } else {
                    if (rep_rooty / corn_dist_up > min_cosine) {
                        cosine = (avg_x - rep_rootx) / dist;
                        dist2 = fabs(((avg_x) / cosine) - dist);
                        percent = dist / (dist + dist2);
                    } else {
                        cosine = (avg_y - rep_rooty) / dist;
                        dist2 = fabs(((avg_y) / cosine) - dist);
                        percent = dist / (dist + dist2);
                    }
                }

                draw_dist(percent, w);
            } else {
                PRINTF("Failed to convert EYE coordinates.\n"); 
            }

            mouse.x = Tx(dx, dy, &w->eyes.t);
            mouse.y = Ty(dx, dy, &w->eyes.t);
            if (!TPointEqual(mouse, w->eyes.mouse)) {
                computePupils(mouse, newpupil);                       /* N */

                xpupil.x = Xx(w->eyes.pupil[0].x, 
                              w->eyes.pupil[0].y, &w->eyes.t);
                xpupil.y = Xy(w->eyes.pupil[0].x, 
                              w->eyes.pupil[0].y, &w->eyes.t);
                xnewpupil.x = Xx(newpupil[0].x, newpupil[0].y, &w->eyes.t);
                xnewpupil.y = Xy(newpupil[0].x, newpupil[0].y, &w->eyes.t);
                if (!XPointEqual(xpupil, xnewpupil)) {
                    if (w->eyes.pupil[0].x != -1000 || 
                        w->eyes.pupil[0].y != -1000) {
                        eyeBall(w, w->eyes.centerGC0, 0);
                    }
                    w->eyes.pupil[0] = newpupil[0];

                    for (i = 0; i < num_colors; i++) {
                        if ( color == dis_colors[i].number ) {
                            eyeBall(w, dis_colors[i].colorGC, 0);
                            break;
                        }
                    }
                }
      
                xpupil.x = Xx(w->eyes.pupil[1].x, 
                              w->eyes.pupil[1].y, &w->eyes.t);
                xpupil.y = Xy(w->eyes.pupil[1].x, 
                              w->eyes.pupil[1].y, &w->eyes.t);
                xnewpupil.x =  Xx(newpupil[1].x, newpupil[1].y, &w->eyes.t);
                xnewpupil.y =  Xy(newpupil[1].x, newpupil[1].y, &w->eyes.t);
                if (!XPointEqual(xpupil, xnewpupil)) {
                    if (w->eyes.pupil[1].x != -1 || w->eyes.pupil[1].y != -1) {
                        eyeBall(w, w->eyes.centerGC1, 1);
                    }
                    w->eyes.pupil[1] = newpupil[1];

                    for (i = 0; i < num_colors; i++) {
                        if (color == dis_colors[i].number) {
                            eyeBall(w, dis_colors[i].colorGC, 1);
                            break;
                        }
                    }
                }
                w->eyes.mouse = mouse;
                w->eyes.update = 0;
            } else {
                if (delays[w->eyes.update + 1] != 0) {
                    ++w->eyes.update;
                }
            }
        }
    }
 
    /* Cast on draw_it() function should not be needed. */
    w->eyes.interval_id =
        XtAddTimeOut(delays[w->eyes.update], 
                     (XtTimerCallbackProc) draw_it, (caddr_t) w);
}


/* Repaint the eyes when they are exposed. */

static void
repaint_window(EyesWidget w)
{
    int i;

    if (XtIsRealized((Widget) w)) {

        /* Repaint the rims of the eyes. */

        for (i = 0; i < num_colors; i++) {
            if (color == dis_colors[i].number) {
                eyeLiner(w, XtWindow(w), dis_colors[i].colorGC,
                         w->eyes.centerGC0, 0);
                eyeLiner(w, XtWindow(w), dis_colors[i].colorGC,
                         w->eyes.centerGC1, 1);
                break;
            }
        }

        /* Compute the pupils' locations. */
        computePupils(w->eyes.mouse, w->eyes.pupil);

        /* Repaint the eyeballs. */
        for (i = 0; i < num_colors; i++) {
            if (color == dis_colors[i].number) {
                eyeBall(w, dis_colors[i].colorGC, 0);
                eyeBall(w, dis_colors[i].colorGC, 1);
                break;
            }
        }
    }
}


/* ARGSUSED */
static Boolean 
SetValues(Widget current, Widget request, Widget new)
{
    return(FALSE);
}


/* Fill in the rims of the eyes. */

static void
eyeLiner(EyesWidget w, Drawable d, GC outgc, GC centergc, int num)
{
    Display *dpy = XtDisplay(w);
  
    TFillArc(dpy, d, outgc, &w->eyes.t,
             EYE_X(num) - EYE_WIDTH / 2.0 - EYE_THICK,
             EYE_Y(num) - EYE_HEIGHT / 2.0 - EYE_THICK,
             EYE_WIDTH + EYE_THICK * 2.0,
             EYE_HEIGHT + EYE_THICK * 2.0,
             90 * 64, 360 * 64);
    if (centergc) {
        TFillArc(dpy, d, centergc, &w->eyes.t,
                 EYE_X(num) - EYE_WIDTH / 2.0,
                 EYE_Y(num) - EYE_HEIGHT / 2.0,
                 EYE_WIDTH, EYE_HEIGHT,
                 90 * 64, 360 * 64);
    }
}


/* Fill in the eyeballs. */

static void
eyeBall(EyesWidget w, GC gc, int num)
{
    Display *dpy = XtDisplay(w);
    Window win = XtWindow(w);

    TFillArc(dpy, win, gc, &w->eyes.t,
             w->eyes.pupil[num].x - BALL_WIDTH / 2.0,
             w->eyes.pupil[num].y - BALL_HEIGHT / 2.0,
             BALL_WIDTH, BALL_HEIGHT,
             90 * 64, 360 * 64);
}
