
/*
 *  UnWindows 1.1.2
 *  dynamag.c
 *
 *  Richard L. Kline
 *  Rensselaer Polytechnic Institute
 *  16 May 1994
 *
 *  Minor revisions by Rich Burridge
 *  Sun Microsytems Inc.
 *  October 2001, January 2002.
 *
 *  This is a major rewrite of the xmag program distributed by 
 *  MIT with X11R5.  It features three modes of magnification.  The
 *  magnifier can be made to follow the mouse pointer around,
 *  displaying a magnified image either in a window that is
 *  "sticky" to the pointer, or in a stationary window.
 *  The magnifier can also be `anchored' to continually magnify a fixed
 *  area of the screen.
 *
 *  The copyright notice of the original program appears below,
 *  though, admittedly, there isn't much of it left.
 *
 *  ---
 *  Modification history:
 *  1.1.3 - October 2002 - see the ChangeLog file.
 *  1.1.1 - October 2001 - see the ChangeLog file.
 *  1.1   - 16 May 94    - added new follow mode without sticky window
 *                         fixed bug in window resizing code
 *                         eliminated nasty lag in response of command
 *                         buttons when magnified area was large
 */

/*
 * xmag - simple magnifying glass application
 *
 * $XConsortium: xmag.c,v 1.29 89/12/10 17:23:58 rws Exp $
 *
 * Copyright 1988 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * Author:  Jim Fulton, MIT X Consortium
 */

#include <stdio.h>
#include <stdlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Scrollbar.h>
#include <Xm/Protocols.h>

#include "corner32.xbm"
#include "corner32m.xbm"
#include "corner64.xbm"
#include "corner64m.xbm"

#define FPRINTF (void) fprintf
#define FSCANF  (void) fscanf
#define SPRINTF (void) sprintf
#define STRCAT  (void) strcat
#define STRCPY  (void) strcpy

#define DEFAULT_BORDER_SIZE     2
#define DEFAULT_CURSOR_SIZE     64
#define DEFAULT_PIX_FONT        "fixed"

#define DEFAULT_MAG_HEIGHT      150
#define DEFAULT_MAG_WIDTH       470
#define DEFAULT_MAGNIFICATION   5
#define DEFAULT_UPDATE          0

#define MOTION_THRESH           4

static char *ProgramName;                /* name of program being executed */
static Display *dpy = NULL;              /* X server being used */
static int screen;                       /* screen being used */
static int pixmap_format;                /* XYPixmap or ZPixmap */
static Colormap cmap;                    /* colormap to use for drawing */
static Window root;                      /* root window of screen */
static GC fillGC = (GC) NULL;            /* for drawing enlargement */
static Cursor llAngle;                   /* custom cursor shape */
static Pixmap pixcursor = (Pixmap) NULL; /* pixmap for larger cursor shape,
                                            if available on display */
static Pixmap pixmask = (Pixmap) NULL;   /* mask pixmap for large cursor */
static Window t = (Window) NULL;

static Atom XA_WM_DELETE_WINDOW;
static XtAppContext appCtx;
static int mag_toggle = False;
static int following  = False;
static int rightnow   = False;
static int mag_height = DEFAULT_MAG_HEIGHT;
static int mag_width  = DEFAULT_MAG_WIDTH;
static int mag        = DEFAULT_MAGNIFICATION;
static int update_int = DEFAULT_UPDATE;
static int src_height = DEFAULT_CURSOR_SIZE;
static int src_width  = DEFAULT_CURSOR_SIZE;
static int display_width;

static XtIntervalId next_timeout;

static int root_x, root_y;      /* pointer's current screen coordinates */
static int old_x = 9999;        /* pointer's previous position */
static int old_y = 9999;
static int anchor_x, anchor_y;  /* lower left corner of magnified screen area */

static char mag_specs_file[128];

static String fallback_resources[] = {
  "*font:         -b&h-lucidatypewriter-bold-r-normal-sans-*-350-*-*-*-*-iso8859-*",
  "*buttons*skipAdjust: True",
  "*Paned*showGrip: False",
  NULL,
};

static Widget toplevel, bigview, follow, onoff, setshell;
static Widget mag_label, mag_scroll, upd_label, upd_scroll, stick_x;
static float mag_scroll_pos, upd_scroll_pos;
static int sticky = TRUE;

static float upd_scroll_values[10] = {
  999.0, 0.25, 0.5, 1.0, 2.0, 3.0, 4.0, 5.0, 10.0, 20.0
};

static char upd_scroll_labels[10][9] = {
  "disabled", "0.25 sec", "0.5 sec", "1 sec", "2 sec", "3 sec",
  "4 sec", "5 sec", "10 sec", "20 sec"
};

static void setup_window(int, int);
static void repaint_image();
static void do_magnify();

static void Quit(Widget, XtPointer, XtPointer); 
static void Follow(Widget, XtPointer, XtPointer);
static void MagnifyOnOff(Widget, XtPointer, XtPointer);
static void Settings(Widget, XtPointer, XtPointer);
static void KillSettings(Widget, XtPointer, XtPointer);
static void SaveSettings(Widget, XtPointer, XtPointer);
static void StickyOnOff(Widget, XtPointer, XtPointer);

static void UpdateMagnifier(XtPointer, XtIntervalId *);
static void LoadSettings(); 
static void UpdateScroll(); 
static void NewMagLevel();


/* Used to respond to DELETE_WINDOW protocol request. */

static void
WindowMessageCB(Widget w, XtPointer client_data, XtPointer call_data)
{
    exit(0);
}


/* Exit procedure. */

/*ARGSUSED*/
static void
Quit(Widget widget, XtPointer closure, XtPointer call_data)
{
    int exitval = (int) call_data;

    if (dpy) {
        XCloseDisplay(dpy);
        dpy = NULL;
    }
    (void) exit(exitval);
}


/* Usage - print out usage message and quit. */

static void
Usage()
{
    static char *help_message[] = {
        "where options include:",
        "    -display displaystring        X server to contact",
        "",
        "It may be abbreviated to -d.",
        "",
        NULL
    }, **cpp;

    FPRINTF(stderr, "usage:\n        %s [-option ...]\n\n", ProgramName);
    for (cpp = help_message; *cpp; cpp++) {
        FPRINTF(stderr, "%s\n", *cpp);
    }
    Quit(NULL, NULL, (XtPointer) 1);
}


/* Follow - make the window follow the cursor around the screen.
 *          Callback function for first command button.
 */

/*ARGSUSED*/
static void
Follow(Widget w, XtPointer closure, XtPointer call_data)
{
    char geom[12];
    XtIntervalId new;

    /* Start following the cursor around. */
    if (!following) {
        following = True;

        /* If we're being sticky, hide the application window. */

        if (sticky) {
            XtUnrealizeWidget(toplevel);
            t = (Window) NULL;

        /* If not, take over the pointer so we can capture the button 
         * press that turns off this mode.
         */

        } else {
            XtGrabPointer(toplevel, False,
                      PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
                      GrabModeAsync, GrabModeAsync, None, llAngle, CurrentTime);
            XtAddRawEventHandler(toplevel, ButtonPressMask | ButtonReleaseMask, 
                                 True, (XtEventHandler) Follow, NULL);
        }

        /* Timeout updates do very bad things while following is on. */

        if (next_timeout) {
            XtRemoveTimeOut (next_timeout);
            next_timeout = (XtIntervalId) 0;

            rightnow = True;
            repaint_image(mag_height, mag_width, mag, 0);
        }

    /* Stop following the cursor and get back to widgets. */

    } else {
        following = False;

        /* If in sticky mode, we must bring back the application widgets. */

        if (sticky) {
            SPRINTF(geom, "+%d+%d", root_x, root_y);
            XtVaSetValues(toplevel, XtNgeometry, geom, NULL);
            XtRealizeWidget(toplevel);

        /* If not, we must return the pointer which we stole while following. */

        } else {
            XtUngrabPointer(toplevel, CurrentTime);
            XtRemoveRawEventHandler(toplevel, 
                                    ButtonPressMask | ButtonReleaseMask,
                                    True, (XtEventHandler)Follow, NULL);
        }

        XSync(dpy, False);
        rightnow = True;
        repaint_image(mag_height, mag_width, mag, 0);

        /* Create a new timeout if update and magnifier are currently on. */

        if (update_int && mag_toggle) {
            new = XtAppAddTimeOut(appCtx, 
                                  (int) (upd_scroll_values[update_int] * 1000),
                                  (XtTimerCallbackProc) UpdateMagnifier, NULL);
        } else {
            new = (XtIntervalId) 0;
        }

        /* If there's already a timeout pending, get rid of it.*/

        if (next_timeout) {
            XtRemoveTimeOut(next_timeout);
        }

        next_timeout = new;
    }
}


/* MagnifyOnOff - toggle the magnifier active/inactive.
 *                Callback function for second command button.
 */

/*ARGSUSED*/
static void
MagnifyOnOff(Widget widget, XtPointer closure, XtPointer call_data)
{
    if (mag_toggle) {
        XtVaSetValues(onoff, XtNlabel, "Resume", NULL);

        /* If there's already a timeout pending, get rid of it. */
        if (next_timeout) {
            XtRemoveTimeOut (next_timeout);
        }
        next_timeout = (XtIntervalId) 0;
    } else {  

        /* Create a new timeout if update is currently on. */

        if (update_int) {
            next_timeout = XtAppAddTimeOut(appCtx,
                                 (int) (upd_scroll_values[update_int] * 1000),
                                 UpdateMagnifier, NULL);
        }

        XtVaSetValues(onoff, XtNlabel, " Stop ", NULL);
        rightnow = True;
        repaint_image(mag_height, mag_width, mag, 0);
    }
    mag_toggle = !mag_toggle;
}


/* Settings - Pop up the settings window. */

/*ARGSUSED*/
static void
Settings(Widget widget, XtPointer closure, XtPointer call_data)
{
    char ch[20];
    Position x, y;

    SPRINTF(ch, "Magnify: X %d", mag);
    XtVaSetValues(mag_label, XtNlabel, ch, NULL);
    mag_scroll_pos = (mag - 2) * 0.125;
    XawScrollbarSetThumb(mag_scroll, mag_scroll_pos, 0.1);

    XtVaSetValues(upd_label, XtNlabel, "Update: disabled", NULL);
    SPRINTF(ch, "Update: %s", upd_scroll_labels[update_int]);
    upd_scroll_pos = update_int * 0.1;
    XawScrollbarSetThumb(upd_scroll, upd_scroll_pos, 0.1);

    if (sticky) {
        XtVaSetValues(stick_x, XtNlabel, "Yes", NULL);
    } else {
        XtVaSetValues(stick_x, XtNlabel, "No", NULL);
    }

    /* Turn off timeouts while this window is up. */
    if (next_timeout) {
        XtRemoveTimeOut (next_timeout);
    }

    XtTranslateCoords(follow, (Position) 0, (Position) 0, &x, &y);
    XtVaSetValues(setshell, XtNx, x, XtNy, y, NULL);
    XtPopup(setshell, XtGrabNonexclusive);
    XtVaSetValues(upd_label, XtNlabel, ch, NULL);
}


/* NewMagLevel - The user has just fiddled with the mag level scrollbar. */

/*ARGSUSED*/
static void
NewMagLevel(Widget w, XtPointer client_data, XtPointer position)
{
    char ch[20];

    if ((int) client_data == 1) {           /* left or right button pressed */
        if ((int) position > 0) {           /* left button was pressed */
            if (mag_scroll_pos >= 0.125) {
                mag_scroll_pos = (int) (mag_scroll_pos * 8) / 8.0 - 0.125;
            } else {
                mag_scroll_pos = 0.0;
            }
        } else {                            /* right button was pressed */
            if (mag_scroll_pos < 0.875) {
                mag_scroll_pos = (int) (mag_scroll_pos * 8) / 8.0 + 0.125;
            } else {
                mag_scroll_pos = 0.875;
            }
        }
    } else {
        mag_scroll_pos = *((float *) position);
    }

    XawScrollbarSetThumb(mag_scroll, mag_scroll_pos, 0.1);
    mag = (int) (mag_scroll_pos * 8) + 2;
    src_height = mag_height / mag;
    src_width  = mag_width  / mag;
    rightnow = True;
    repaint_image(mag_height, mag_width, mag, 0);
    SPRINTF(ch, "Magnify: X %d", mag);
    XtVaSetValues(mag_label, XtNlabel, ch, NULL);
}


/* UpdateScroll - The user has just fiddled with the update scrollbar. */

/*ARGSUSED*/
static void
UpdateScroll(Widget w, XtPointer client_data, XtPointer position)
{
    char ch[20];

    if ((int) client_data == 1) {            /* left or right button pressed */
        if ((int) position > 0) {            /* left button was pressed */
            if (upd_scroll_pos >= 0.1) {
                upd_scroll_pos = (int) (upd_scroll_pos * 10) / 10.0 - 0.1;
            } else {
                upd_scroll_pos = 0.0;
            }
        } else {                             /* right button was pressed */
            if (upd_scroll_pos < 0.899) {
                upd_scroll_pos = (int) (upd_scroll_pos * 10) / 10.0 + 0.1;
            } else {
                upd_scroll_pos = 0.9;
            }
        }
    } else {
        upd_scroll_pos = *((float *) position);
    }

    XawScrollbarSetThumb(upd_scroll, upd_scroll_pos, 0.1);
    update_int = (int) (upd_scroll_pos * 10);
    SPRINTF(ch, "Update: %s", upd_scroll_labels[update_int]);
    XtVaSetValues(upd_label, XtNlabel, ch, NULL);
}


/*ARGSUSED*/
static void
UpdateMagnifier(XtPointer x, XtIntervalId *interval)
{
    XtIntervalId new;

    if (mag_toggle) {
        rightnow = True;
        repaint_image(mag_height, mag_width, mag, 0);

        /* Create a new timeout if update is currently on. */
        if (update_int) {
            new = XtAppAddTimeOut(appCtx,
                                  (int) (upd_scroll_values[update_int]*1000),
                                  UpdateMagnifier, NULL);
        } else {
            new = (XtIntervalId) 0;
        }

        /* If there's already a timeout pending, get rid of it. */
        if (next_timeout) {
            XtRemoveTimeOut (next_timeout);
        }
        next_timeout = new;
    }
}


/* KillSettings - User's done with the settings window. */

/*ARGSUSED*/
static void
KillSettings(Widget widget, XtPointer closure, XtPointer call_data)
{
    /* Create a new timeout if update and magnifier are currently on. */

    if (update_int && mag_toggle) {
        XtAppAddTimeOut(appCtx, (int) (upd_scroll_values[update_int] * 1000),
                        UpdateMagnifier, NULL);
    }

    XtPopdown(setshell);
}


/* LoadSettings - Read in the user's preferences from the appropriate
 *                file.  Set defaults if the file is incomplete, has
 *                bad values, or is altogether missing.
 */

static void
LoadSettings()
{
    FILE * fp;
    char *home, line[80];

    if ((home = getenv("HOME")) != NULL) {
        SPRINTF(mag_specs_file, "%s/.un_xeyes", home);
    } else {
        STRCPY(mag_specs_file, "/.dynamag");
    }
    fp = fopen(mag_specs_file, "r");
    if (fp) {
        FSCANF(fp, "%s %d", line, &mag_height);
        FSCANF(fp, "%s %d", line, &mag_width);
        FSCANF(fp, "%s %d", line, &mag);
        FSCANF(fp, "%s %d", line, &update_int);
        FSCANF(fp, "%s %d", line, &sticky);
    } else {
        mag_height = DEFAULT_MAG_HEIGHT;
        mag_width  = DEFAULT_MAG_WIDTH;
        mag        = DEFAULT_MAGNIFICATION;
        update_int = DEFAULT_UPDATE;
        sticky     = FALSE;
    }

  src_height = mag_height / mag;
  src_width  = mag_width / mag;
}


/* SaveSettings - Save the current configuration into the user's
 *                preferences file.
 */

/*ARGSUSED*/
static void
SaveSettings(Widget widget, XtPointer closure, XtPointer call_data)
{
    FILE * fp = fopen(mag_specs_file, "w");

    if (fp) {
        FPRINTF(fp, "magnifier_height    %d\n", mag_height);
        FPRINTF(fp, "magnifier_width    %d\n", mag_width);
        FPRINTF(fp, "magnification        %d\n", mag);
        FPRINTF(fp, "update_interval    %d\n", update_int);
        FPRINTF(fp, "sticky        %d\n", sticky);
    } else {
        FPRINTF(stderr, "Unable to save settings to '%s'!\n", mag_specs_file);
    }
}


/* StickyOnOff - Switches behavior of mobile mode between mobile and
 *               stationary magnification window
 */

/*ARGSUSED*/
static void
StickyOnOff(Widget widget, XtPointer closure, XtPointer call_data)
{
    if (sticky) {
        sticky = FALSE;
        XtVaSetValues(stick_x, XtNlabel, "No", NULL);
    } else {
        sticky = TRUE;
        XtVaSetValues(stick_x, XtNlabel, "Yes", NULL);
    }
}


int
main(int argc, char **argv)
{
    int i;                           /* iterator variable */
    char *displayname = NULL;        /* name of X server to contact */

    ProgramName = argv[0];

    for (i = 1; i < argc; i++) {
        char *arg = argv[i];

        if (arg[0] == '-') {         /* process command line switches */
            switch (arg[1]) {
                case 'd':            /* -display displaystring */
                    if (++i >= argc) {
                        Usage();
                    }
                    displayname = argv[i];
                    continue;
                default:
                    Usage();
            }
        } else {                     /* any command line parameters */
            Usage();
        }
    }

    toplevel = XtAppInitialize(&appCtx, "Dynamag", NULL, 0,
                               &argc, argv, fallback_resources, NULL, 0);

    dpy = XtDisplay(toplevel);
    
    if (!dpy) {
        FPRINTF(stderr, "%s:  unable to open display \"%s\".\n",
                ProgramName, XDisplayName (displayname));
        Quit(NULL, NULL, (XtPointer) 1);
    }

    display_width  = DisplayWidth(dpy, DefaultScreen(dpy));

    root   = DefaultRootWindow(dpy);
    screen = DefaultScreen(dpy);
    cmap   = DefaultColormap(dpy, screen);
    pixmap_format = (DisplayPlanes(dpy, screen) > 1 ? ZPixmap : XYPixmap);

    XSelectInput(dpy, root, 
                 PointerMotionMask | KeyPressMask | StructureNotifyMask);

    LoadSettings();
    setup_window(mag_height, mag_width);
    XtRealizeWidget(toplevel);

    XA_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(toplevel, XA_WM_DELETE_WINDOW,
                            WindowMessageCB, (caddr_t) 0);

    /* Turn it on */
    mag_toggle = False;
    MagnifyOnOff(NULL, NULL, NULL);

    /* Create fancy cursor for mobile mode */

    {
        unsigned int wout, hout, hotx = 0, hoty = 0;
        XColor black, white;
        Drawable d = XtWindow(toplevel);

        /* Try to make a 64x64 cursor if the display will allow it... if
         * not, try for 32x32.
         */

        XQueryBestCursor(dpy, d, 64, 64, &wout, &hout);
        XParseColor(dpy, cmap, "white", &white);
        XParseColor(dpy, cmap, "black", &black);
        if (wout > 63 && hout > 63) {
            pixcursor = XCreateBitmapFromData(dpy, XtWindow(toplevel),
                                              (const char *) corner64_bits,
                                              corner64_width, corner64_height);
            pixmask = XCreateBitmapFromData(dpy, XtWindow(toplevel),
                                            (const char *) corner64m_bits,
                                            corner64m_width, corner64m_height);
            hotx = corner64_x_hot;
            hoty = corner64_y_hot;
        } else if (wout > 31 && hout > 31) {
            pixcursor = XCreateBitmapFromData(dpy, XtWindow(toplevel),
                                              (const char *) corner32_bits,
                                              corner32_width, corner32_height);
            pixmask = XCreateBitmapFromData(dpy, XtWindow(toplevel),
                                            (const char *) corner32m_bits, 
                                            corner32m_width, corner32m_height);
            hotx = corner32_x_hot;
            hoty = corner32_y_hot;
        }

        /* If neither of these sizes were allowed, or if the files couldn't
         * be properly read, use the built-in cursor shape.
         */

        if (!pixcursor) {
            llAngle = XCreateFontCursor(dpy, XC_ll_angle);
        } else {
            llAngle = XCreatePixmapCursor(dpy, pixcursor, pixmask, 
                                          &white, &black, hotx, hoty);
            XFreePixmap(dpy, pixcursor);
            XFreePixmap(dpy, pixmask);
        }
    }

    XSync(dpy, False);
    do_magnify();

    Quit(NULL, NULL, (XtPointer) 0);

    return(0);
}


/* 
 * Set up all the fine widgets that make up the application.
 * Figure out the initial settings for the magnifier.
 */

static void
setup_window(int height, int width)
{
    /* Parent window stuff */
    Widget mom, temp;

    /* Pop-up window stuff */
    Dimension ht;
    Widget setform;

    mom = XtVaCreateManagedWidget("the_form", panedWidgetClass,
                                  toplevel, NULL);

    bigview = XtVaCreateManagedWidget("bigview", widgetClass, mom,
                                      XtNheight, height, 
                                      XtNwidth, width,
                                      XtNbackground, BlackPixel(dpy, screen), 
                                      NULL);

    mom = XtVaCreateManagedWidget("buttons", formWidgetClass, mom, NULL);

    follow = XtVaCreateManagedWidget("Follow", commandWidgetClass, mom, NULL);
    XtAddCallback(follow, XtNcallback, Follow, NULL);

    onoff = XtVaCreateManagedWidget("onoff", commandWidgetClass, mom,
                                    XtNfromHoriz, follow, 
                                    XtNlabel, " Stop ", 
                                    NULL);
    XtAddCallback(onoff, XtNcallback, MagnifyOnOff, NULL);

    temp = XtVaCreateManagedWidget("Settings", commandWidgetClass, mom,
                                   XtNfromHoriz, onoff,
                                   NULL);
    XtAddCallback(temp, XtNcallback, Settings, NULL);

    temp = XtVaCreateManagedWidget("Quit", commandWidgetClass, mom, 
                                   XtNfromHoriz, temp, 
                                   NULL);
    XtAddCallback(temp, XtNcallback, Quit, NULL);

    /* Set up the settings pop-up window */

    if (!setshell) {
        setshell = XtVaCreatePopupShell("setshell",
                                        transientShellWidgetClass, toplevel,
                                        XtNtitle, "dynamag settings",
                                        NULL);
    
        setform = XtVaCreateManagedWidget("setform", formWidgetClass, setshell,
                                          NULL);
    
        mag_label = XtVaCreateManagedWidget("mag_label", 
                                            labelWidgetClass, setform,
                                            XtNlabel, "Magnify: X 2", 
                                            XtNborderWidth, 0,
                                            NULL);

        upd_label = XtVaCreateManagedWidget("upd_label",
                                            labelWidgetClass, setform, 
                                            XtNlabel, "Update: disabled",
                                            XtNjustify, XtJustifyLeft,
                                            XtNfromVert, mag_label,
                                            XtNborderWidth, 0,
                                            NULL);

        /* Get height of label (which was automatically chosen based on the
         * label font size) to use in setting height of companion widgets.
         */

        XtVaGetValues(mag_label, XtNheight, &ht, NULL);

        mag_scroll = XtVaCreateManagedWidget("mag_scroll",
                                             scrollbarWidgetClass, setform,
                                             XtNorientation, XtorientHorizontal,
                                             XtNfromHoriz, upd_label, 
                                             XtNheight, ht,
                                             XtNwidth, 300, 
                                             XtNminimumThumb, 40, 
                                             NULL);
        XtAddCallback(mag_scroll, XtNscrollProc, NewMagLevel, (XtPointer) 1);
        XtAddCallback(mag_scroll, XtNjumpProc, NewMagLevel, (XtPointer) 2);

        upd_scroll = XtVaCreateManagedWidget("upd_scroll", 
                                             scrollbarWidgetClass, setform, 
                                             XtNorientation, XtorientHorizontal,
                                             XtNfromVert, mag_label,
                                             XtNfromHoriz, upd_label, 
                                             XtNheight, ht,
                                             XtNwidth, 300, 
                                             XtNminimumThumb, 40,
                                             NULL);
        XtAddCallback(upd_scroll, XtNscrollProc, UpdateScroll, (XtPointer) 1);
        XtAddCallback(upd_scroll, XtNjumpProc, UpdateScroll, (XtPointer) 2);

        temp = XtVaCreateManagedWidget("stick_label",
                                       labelWidgetClass, setform,
                                       XtNlabel, "Sticky window",
                                       XtNjustify, XtJustifyLeft,
                                       XtNfromVert, upd_label,
                                       XtNborderWidth, 0,
                                       NULL);

        stick_x = XtVaCreateManagedWidget("stick_x", 
                                          commandWidgetClass, setform,
                                          XtNlabel, "Yes", 
                                          XtNfromVert, upd_label,
                                          XtNfromHoriz, upd_label, 
                                          XtNresize, False,
                                          NULL);
        XtAddCallback(stick_x, XtNcallback, StickyOnOff, NULL);

        temp = XtVaCreateManagedWidget("setsave", 
                                       commandWidgetClass, setform,
                                       XtNfromVert, stick_x,
                                       XtNlabel, "Save settings",
                                       NULL);
        XtAddCallback(temp, XtNcallback, SaveSettings, NULL);

        temp = XtVaCreateManagedWidget("setdone", 
                                       commandWidgetClass, setform,
                                       XtNfromVert, stick_x, 
                                       XtNfromHoriz, temp,
                                       XtNlabel, "Done",
                                       NULL);
        XtAddCallback (temp, XtNcallback, KillSettings, NULL);
    }
}


/* Display the magnifying glass and follow the pointer.  */

static void 
do_magnify()
{
    XColor white;
    int window_width = 0;
    int window_height = 0;
    Position x, y;
  
    XParseColor(dpy, cmap, "White", &white);
    XAllocColor(dpy, cmap, &white);

    anchor_x = 0;
    anchor_y = src_height;
  
    while (1) {            /* do forever */
        XEvent event;
        XEvent peek;
        XMotionEvent xme;
    
        XtAppNextEvent(appCtx, &event);
        XtDispatchEvent(&event);
    
        /* End pointer-following mode if mouse button is pressed
         * inside magnifier window.
         */
    
        if (following && event.type == ButtonPress) {
            XDestroyWindow(dpy, t);
            t = (Window) NULL;
            Follow((Widget) t, NULL, NULL);
    
            /* Remember where the window ended up, since the magnifier
             * is still on.
             */

            anchor_x = root_x; 
            anchor_y = root_y;
            old_x = old_y = 9999;
        }
    
        /* Figure out new dimensions for magnifier if window has
         * been resized.
         */

        if (event.type == ConfigureNotify) {
            XConfigureEvent *xce = &(event.xconfigure);
            if (xce->width != window_width || xce->height != window_height) {
                XtVaGetValues(bigview, XtNwidth, &x, XtNheight, &y, NULL);
                mag_width = x;
                mag_height = y;
                src_width = mag_width / mag;
                src_height = mag_height / mag;
                window_width = xce->width;
                window_height = xce->height;
            }
        }

        /* We can get up to five of these at once, depending on how many
         * of the command buttons are uncovered as well as the image area. 
         * But we only want to redraw once.
         */

        if (event.type == Expose) {
            repaint_image(mag_height, mag_width, mag, 0);
            if (QLength (dpy)) {
                XPeekEvent(dpy, &peek);
                while (QLength (dpy) && peek.type == Expose) {
                    XtAppNextEvent(appCtx, &event);
                    XtDispatchEvent(&event);

                    /* If queue is empty, XPeekEvent will block. We don't 
                     * want this.
                     */

                    if (QLength(dpy)) {
                        XPeekEvent(dpy, &peek);
                    }
                }
            }
        }

        /* Redisplay if pointer is moved */    

        if (event.type == MotionNotify) {
      
            /* Look through events to find the most recent MotionNotify event,
             * so we can use the coordinates given there as the most current
             * pointer position.  We can gloss over all event types except
             * the ones above, ButtonPress, Expose, and ConfigureNotify.
             */
      
            xme = event.xmotion;
            if (QLength (dpy)) {
                XPeekEvent(dpy, &peek);
                while (QLength(dpy) > 0 && 
                       peek.type != Expose && 
                       peek.type != ConfigureNotify && 
                       peek.type != ButtonPress) {
                    XtAppNextEvent (appCtx, &event);
                    XtDispatchEvent (&event);
                    if (event.type == MotionNotify) {
                        xme = event.xmotion;
                    }

                    /* If queue is empty, XPeekEvent will block.  We don't 
                     * want this.
                     */

                    if (QLength(dpy)) {
                        XPeekEvent(dpy, &peek);
                    }
                }
            }

            /* Redraw the magnifying glass if:
             *
             * 1 - it is toggled on
             * 2 - the pointer moved more than MOTION_THRESH pixels on
             *     either axis.
             */

            if ((mag_toggle || following) && 
                (abs(xme.x_root - old_x) > MOTION_THRESH || 
                 abs(xme.y_root - old_y) > MOTION_THRESH)) {
                repaint_image(mag_height, mag_width, mag, xme);
            }
        }
    }
}


/* Copy the appropriate screen area and fill in the magnifying glass
 *
 *  NOTE: I tried XDrawLine in place of XFillRectangle, but it was
 *  no faster and produced some flickering and some incorrect drawing.
 */

static void
repaint_image(int height, int width, int magnification, XMotionEvent xme)
{
    XImage *image;
    int r_x, r_y, r_width, r_height;
    XGCValues gcv;
    int row, column;
    int maxrow, maxcolumn;
    unsigned long pixel;

    XSetWindowAttributes attr;

    attr.save_under = True;
    attr.override_redirect = True;
    attr.border_pixel = BlackPixel (dpy, screen);
    attr.background_pixel = WhitePixel (dpy, screen);

    /* One-shot conditions ("rightnow" is TRUE):
     * - "follow" has just been turned on
     * - magnify level has been changed
     * - update time-out has just occurred.
     */
          
    if (following) {
        if (rightnow) {
            rightnow = False;
            root_x = anchor_x;
            root_y = anchor_y;
        } else {
            old_x = root_x = xme.x_root;
            old_y = root_y = xme.y_root;
    
            if (sticky && t) {
                XDestroyWindow(dpy, t);
            }
        }
    }

    if (root_x <= 0) {
        root_x = 0;
    }
    if (root_y <= 0) {
        root_y = 0;
    }
    
    if (root_y - src_height <= 0) {
        if (root_x + src_width > display_width) {
            image = XGetImage(dpy, root, display_width-src_width, 0,
                              src_width, src_height, AllPlanes, pixmap_format);
        } else {
            image = XGetImage(dpy, root, root_x, 0,
                              src_width, src_height, AllPlanes, pixmap_format);
        }
    } else {
        if (root_x + src_width > display_width) {
            image = XGetImage(dpy, root, display_width-src_width, 
                              root_y - src_height, 
                              src_width, src_height, AllPlanes, pixmap_format);
        } else {
            image = XGetImage(dpy, root, root_x, root_y - src_height,
                              src_width, src_height, AllPlanes, pixmap_format);
        }
    }
    
    /* If the magnifier is currently following the cursor,
     * it is its own unmanaged window and must be created.
     * If not, then the "bigdraw" box widget is where the
     * magnified image will be drawn.
     */

    if (following && sticky) {
        r_x = src_width * magnification;
        r_y = src_height * magnification;
        if (root_y - mag_height <= 0) {
            if (root_x + mag_width - MOTION_THRESH > display_width) {
                t = XCreateWindow(dpy, root, display_width - r_x, 0,
                                  r_x, r_y, 2, 8, InputOutput, CopyFromParent,
                                  CWSaveUnder | CWBorderPixel |
                                  CWOverrideRedirect, &attr);
            } else {
                t = XCreateWindow(dpy, root, root_x-MOTION_THRESH, 0,
                                  r_x, r_y, 2, 8, InputOutput, CopyFromParent,
                                  CWSaveUnder | CWBorderPixel |
                                  CWOverrideRedirect, &attr);
            }
        } else {
            if (root_x + r_x - MOTION_THRESH> display_width) {
                t = XCreateWindow(dpy, root, display_width - r_x, 
                                  root_y - r_y+MOTION_THRESH, r_x, r_y,
                                  2, 8, InputOutput, CopyFromParent,
                                  CWSaveUnder | CWBorderPixel |
                                  CWOverrideRedirect, &attr);
            } else {
                t = XCreateWindow(dpy, root, root_x-MOTION_THRESH,
                                  root_y - r_y+MOTION_THRESH, r_x, r_y,
                                  2, 8, InputOutput, CopyFromParent,
                                  CWSaveUnder | CWBorderPixel |
                                  CWOverrideRedirect, &attr);
            }
        }
        XSelectInput(dpy, t, ButtonPressMask);
        XMapWindow(dpy, t);
    } else {
        t = XtWindow(bigview);
    }
          
    gcv.function = GXcopy;
    gcv.plane_mask = AllPlanes;
    gcv.fill_style = FillSolid;
    
    fillGC = XCreateGC(dpy, t, (GCFunction | GCPlaneMask | GCFillStyle), &gcv);

    gcv.foreground = BlackPixel(dpy, screen);
    XChangeGC(dpy, fillGC, GCForeground, &gcv);

    maxrow    = ((height + magnification - 1) / magnification) - 1;
    maxcolumn = ((width  + magnification - 1) / magnification) - 1;

    /* Clip to image size. */

    if (maxcolumn >= image->width) {
        maxcolumn = image->width - 1;
    }
    if (maxrow >= image->height) {
        maxrow = image->height - 1;
    }

    /* Repaint the image; be somewhat clever by only drawing when the pixel
     * changes or when we run out of lines.  
     */

    r_y = 0;                         /* to start */
    r_height = magnification;        /* always doing just 1 line */

    /* Iterate over scanlines (rows). */

    for (row = 0; row <= maxrow; row++) {
        r_x = 0;                     /* start at beginning */
        r_width = magnification;

        /* Iterate over pixels in scanline (columns). */

        gcv.foreground = XGetPixel(image, 0, row);
        for (column = 1; column <= maxcolumn; column++) {
            pixel = XGetPixel(image, column, row);
            if (pixel != gcv.foreground) {
                XChangeGC(dpy, fillGC, GCForeground, &gcv);
                XFillRectangle(dpy, t, fillGC, r_x, r_y, r_width, r_height);
                gcv.foreground = pixel;
                r_x += r_width;
                r_width = 0;
            }
            r_width += magnification;
        }
        XChangeGC(dpy, fillGC, GCForeground, &gcv);
        XFillRectangle(dpy, t, fillGC, r_x, r_y, r_width, r_height);
        r_y += magnification;                /* Advance to next scanline */
    }
    
    XDestroyImage(image);
}
